#
# testing rexpy with hypothesis
#
# see hypothesis.works
#
# requires hypothesis to be installed, via "pip install hypothesis".
#

import unittest

from tdda.rexpy.relib import re
from tdda.rexpy import Extractor

try:
    from hypothesis import given, settings, HealthCheck
    from hypothesis.strategies import text, integers, composite

    @composite
    def list_of_strings(draw):
        return [draw(text(min_size=1, average_size=70))
                for i in range(draw(integers(min_value=0, max_value=100)))]

    class TestRexpyHypothetically(unittest.TestCase):
        @given(list_of_strings())
        @settings(suppress_health_check=[HealthCheck.too_slow])
        def test_matches(self, strings):
            x = Extractor(strings)
            for rex in x.results.rex:
                r = re.compile(rex, flags=re.DOTALL|re.UNICODE)
                matches = [r.match(e) for e in strings]
                self.assertTrue(any(matches))

except ImportError:
    pass


if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-
"""
TDDA constraint discovery and verification is provided for a number
of DB-API (PEP-0249) compliant databases, and also for a number of other
(NoSQL) databases.

The top-level functions are:

    :py:func:`discover_db_table`:
        Discover constraints from a single database table.

    :py:func:`verify_db_table`:
        Verify (check) a single database table, against a set of previously
        discovered constraints.

    :py:func:`detect_db_table`:
        Verify (check) a single database table, against a set of previously
        discovered constraints.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import sys

from tdda.constraints.base import (
    DatasetConstraints,
    Verification,
)
from tdda.constraints.baseconstraints import (
    BaseConstraintCalculator,
    BaseConstraintDetector,
    BaseConstraintVerifier,
    BaseConstraintDiscoverer,
    MAX_CATEGORIES,
)

from tdda.constraints.db.drivers import DatabaseHandler
from tdda import rexpy

if sys.version_info[0] >= 3:
    long = int


class DatabaseConstraintCalculator(BaseConstraintCalculator):
    def __init__(self, tablename, testing=False):
        self.tablename = tablename
        self.testing = testing

    def is_null(self, value):
        return self.db_value_is_null(value)

    def to_datetime(self, value):
        return self.db_value_to_datetime(value)

    def column_exists(self, colname):
        return self.canon(colname) in self.get_column_names()

    def get_column_names(self):
        return self.get_database_column_names(self.tablename)

    def get_nrecords(self):
        return self.get_database_nrows(self.tablename)

    def types_compatible(self, x, y, colname):
        return types_compatible(x, y, colname if not self.testing else None)

    def calc_min(self, colname):
        return self.get_database_min(self.tablename, colname)

    def calc_max(self, colname):
        return self.get_database_max(self.tablename, colname)

    def calc_min_length(self, colname):
        return self.get_database_min_length(self.tablename, colname)

    def calc_max_length(self, colname):
        return self.get_database_max_length(self.tablename, colname)

    def calc_tdda_type(self, colname):
        return self.get_database_column_type(self.tablename, colname)

    def calc_null_count(self, colname):
        return self.get_database_nnull(self.tablename, colname)

    def calc_non_null_count(self, colname):
        return self.get_database_nnonnull(self.tablename, colname)

    def calc_nunique(self, colname):
        return self.get_database_nunique(self.tablename, colname)

    def calc_unique_values(self, colname, include_nulls=True):
        return self.get_database_unique_values(self.tablename, colname,
                                               include_nulls=include_nulls)

    def calc_non_integer_values_count(self, colname):
        raise Exception('database should not require non_integer_values_count')

    def calc_all_non_nulls_boolean(self, colname):
        raise Exception('database should not require all_non_nulls_boolean')

    def find_rexes(self, colname, values=None):
        if not values:
            values = self.get_database_unique_values(self.tablename, colname)
        return rexpy.extract(values)

    def calc_rex_constraint(self, colname, constraint, detect=False):
        return not self.get_database_rex_match(self.tablename, colname,
                                               constraint.value)


class DatabaseConstraintDetector(BaseConstraintDetector):
    """
    No-op implementation of the Constraint Detector methods for
    databases.
    """
    def __init__(self, tablename):
        pass


class DatabaseConstraintVerifier(DatabaseConstraintCalculator,
                                 DatabaseConstraintDetector,
                                 BaseConstraintVerifier,
                                 DatabaseHandler):
    """
    A :py:class:`DatabaseConstraintVerifier` object provides methods
    for verifying every type of constraint against a single database table.
    """
    def __init__(self, dbtype, db, tablename, epsilon=None,
                 type_checking='strict', testing=False):
        """
        Inputs:

            *dbtype*:
                    Type of database.
            *db*:
                    A DB-API database connection object (for example, as
                    obtained from a call to pgdb.connect() for PostGreSQL
                    or as a call to MySQLdb.connect() for MySQL).
            *tablename*:
                    A table name, referring to a table that exists in the
                    database and is accessible. It can either be a simple
                    name, or a schema-qualified name of the form `schema.name`.
        """
        DatabaseConstraintCalculator.__init__(self, tablename, testing)
        DatabaseConstraintDetector.__init__(self, tablename)
        BaseConstraintVerifier.__init__(self, epsilon=epsilon,
                                        type_checking=type_checking)
        DatabaseHandler.__init__(self, dbtype, db)


class DatabaseVerification(Verification):
    """
    A :py:class:`DatabaseVerification` object is the variant of
    the :py:class:`tdda.constraints.base.Verification` object used for
    verification of constraints on a database table.
    """
    def __init__(self, *args, **kwargs):
        Verification.__init__(self, *args, **kwargs)


class DatabaseConstraintDiscoverer(DatabaseConstraintCalculator,
                                   BaseConstraintDiscoverer,
                                   DatabaseHandler):
    """
    A :py:class:`DatabaseConstraintDiscoverer` object is used to discover
    constraints on a single database table.
    """
    def __init__(self, dbtype, db, tablename, inc_rex=False):
        DatabaseConstraintCalculator.__init__(self, tablename)
        BaseConstraintDiscoverer.__init__(self, inc_rex=inc_rex)
        DatabaseHandler.__init__(self, dbtype, db)
        self.tablename = tablename


def types_compatible(x, y, colname):
    """
    Returns boolean indicating whether the coarse_type of *x* and *y* are
    the same, for scalar values. The int and long types are considered to
    be the same.

    For databases, coarse types are pretty much the same as the column types,
    except that different sizes of integer are all considered to be ints.

    If *colname* is provided, and the check fails, a warning is issued
    to stderr.
    """
    tx = int if type(x) is long else type(x)
    ty = int if type(y) is long else type(y)
    ok = tx == ty
    if not ok and colname:
        print('Warning: Failing incompatible types constraint for field %s '
              'of type %s.\n(Constraint value %s of type %s.)'
              % (colname, type(x), y, type(y)), file=sys.stderr)
    return ok


def verify_db_table(dbtype, db, tablename, constraints_path, epsilon=None,
                    type_checking='strict', testing=False, **kwargs):
    """
    Verify that (i.e. check whether) the database table provided
    satisfies the constraints in the JSON .tdda file provided.

    Mandatory Inputs:

        *dbtype*:
                            Type of database.
        *db*:
                            A database object
        *tablename*:
                            A database table name, to be checked.

        *constraints_path*:
                            The path to a JSON .tdda file (possibly
                            generated by the discover_constraints
                            function, below) containing constraints
                            to be checked.

    Optional Inputs:

        *epsilon*:
                            When checking minimum and maximum values
                            for numeric fields, this provides a
                            tolerance. The tolerance is a proportion
                            of the constraint value by which the
                            constraint can be exceeded without causing
                            a constraint violation to be issued.

                            For example, with epsilon set to 0.01 (i.e. 1%),
                            values can be up to 1% larger than a max constraint
                            without generating constraint failure,
                            and minimum values can be up to 1% smaller
                            that the minimum constraint value without
                            generating a constraint failure. (These
                            are modified, as appropriate, for negative
                            values.)

                            If not specified, an *epsilon* of 0 is used,
                            so there is no tolerance.

                            NOTE: A consequence of the fact that these
                            are proportionate is that min/max values
                            of zero do not have any tolerance, i.e.
                            the wrong sign always generates a failure.

        *type_checking*:
                            'strict' or 'sloppy'. For databases (unlike
                            Pandas DataFrames), this defaults to 'strict'.

                            If this is set to sloppy, a database "real"
                            column c will only be allowed to satisfy a
                            an "int" type constraint.

        *report*:
                            'all' or 'fields'.
                            This controls the behaviour of the
                            :py:meth:`~DatabaseVerification.__str__`
                            method on the resulting
                            :py:class:`~DatabaseVerification`
                            object (but not its content).

                            The default is 'all', which means that
                            all fields are shown, together with the
                            verification status of each constraint
                            for that field.

                            If report is set to 'fields', only fields for
                            which at least one constraint failed are shown.

                            NOTE: The method also accepts two further
                            parameters to control (not yet implemented)
                            behaviour. 'constraints', will be used to
                            indicate that only failing constraints for
                            failing fields should be shown.
                            'one_per_line' will indicate that each constraint
                            failure should be reported on a separate line.

        *testing*:
                            Boolean flag. Should only be set to ``True``
                            when being run as part of an automated test.
                            It suppresses type-compatibility warnings.

    Returns:

        :py:class:`~DatabaseVerification` object.

        This object has attributes:

        - *passed*      --- Number of passing constriants
        - *failures*    --- Number of failing constraints

    Example usage::

        import pgdb
        from tdda.constraints import verify_db_table

        dbspec = 'localhost:databasename:username:password'
        tablename = 'schemaname.tablename'
        db = pgdb.connect(dbspec)
        v = verify_db_table('postgres' db, tablename, 'myconstraints.tdda')

        print('Constraints passing:', v.passes)
        print('Constraints failing: %d\\n' % v.failures)
        print(str(v))
    """
    dbv = DatabaseConstraintVerifier(dbtype, db, tablename, epsilon=epsilon,
                                     type_checking=type_checking,
                                     testing=testing)
    if not dbv.check_table_exists(tablename):
        print('No table %s' % tablename, file=sys.stderr)
        sys.exit(1)
    constraints = DatasetConstraints(loadpath=constraints_path)
    return dbv.verify(constraints,
                      VerificationClass=DatabaseVerification, **kwargs)


def detect_db_table(dbtype, db, tablename, constraints_path, epsilon=None,
                    type_checking='strict', testing=False, **kwargs):
    """
    Detect failures from verification of constraints
    """
    raise NotImplementedError('Detection is not implemented (yet) '
                              'for databases.')


def discover_db_table(dbtype, db, tablename, inc_rex=False):
    """
    Automatically discover potentially useful constraints that characterize
    the database table provided.

    Input:

        *dbtype*:
            Type of database.
        *db*:
            a database object
        *tablename*:
            a table name

    Possible return values:

    -  :py:class:`~tdda.constraints.base.DatasetConstraints` object
    -  ``None``    --- (if no constraints were found).

    This function goes through each column in the table and, where
    appropriate, generates constraints that describe (and are satisified
    by) this dataframe.

    Assuming it generates at least one constraint for at least one field
    it returns a :py:class:`tdda.constraints.base.DatasetConstraints` object.

    This includes a 'fields' attribute, keyed on the column name.

    The returned :py:class:`~tdda.constraints.base.DatasetConstraints` object
    includes a :py:meth:`~tdda.constraints.base.DatasetContraints.to_json`
    method, which converts the constraints into JSON for saving as a tdda
    constraints file. By convention, such JSON files use a '.tdda'
    extension.

    The JSON constraints file can be used to check whether other datasets
    also satisfy the constraints.

    The kinds of constraints (potentially) generated for each field (column)
    are:

        *type*:
                the (coarse, TDDA) type of the field. One of
                'bool', 'int', 'real', 'string' or 'date'.


        *min*:
                for non-string fields, the minimum value in the column.
                Not generated for all-null columns.

        *max*:
                for non-string fields, the maximum value in the column.
                Not generated for all-null columns.

        *min_length*:
                For string fields, the length of the shortest string(s)
                in the field.

        *max_length*:
                For string fields, the length of the longest string(s)
                in the field.

        *sign*:
                If all the values in a numeric field have consistent sign,
                a sign constraint will be written with a value chosen from:

                - positive     --- For all values *v* in field: `v > 0`
                - non-negative --- For all values *v* in field: `v >= 0`
                - zero         --- For all values *v* in field: `v == 0`
                - non-positive --- For all values *v* in field: `v <= 0`
                - negative     --- For all values *v* in field: `v < 0`
                - null         --- For all values *v* in field: `v is null`

        *max_nulls*:
                The maximum number of nulls allowed in the field.

                - If the field has no nulls, a constraint
                  will be written with max_nulls set to zero.
                - If the field has a single null, a constraint will
                  be written with max_nulls set to one.
                - If the field has more than 1 null, no constraint
                  will be generated.

        *no_duplicates*:
                For string fields (only, for now), if every
                non-null value in the field is different,
                this constraint will be generated (with value ``True``);
                otherwise no constraint will be generated. So this constraint
                indicates that all the **non-null** values in a string
                field are distinct (unique).

        *allowed_values*:
                 For string fields only, if there are
                 :py:const:`MAX_CATEGORIES` or fewer distinct string
                 values in the dataframe, an AllowedValues constraint
                 listing them will be generated.
                 :py:const:`MAX_CATEGORIES` is currently "hard-wired" to 20.

    Example usage::

        import pgdb
        from tdda.constraints import discover_db_table

        dbspec = 'localhost:databasename:username:password'
        tablename = 'schemaname.tablename'
        db = pgdb.connect(dbspec)
        constraints = discover_db_table('postgres', db, tablename)

        with open('myconstraints.tdda', 'w') as f:
            f.write(constraints.to_json())

    """
    disco = DatabaseConstraintDiscoverer(dbtype, db, tablename,
                                         inc_rex=inc_rex)
    if not disco.check_table_exists(tablename):
        print('No table %s' % tablename, file=sys.stderr)
        sys.exit(1)
    return disco.discover()





# -*- coding: utf-8 -*-

"""
Common database library support
"""

from __future__ import division
from __future__ import print_function

import datetime
import getpass
import json
import os
import re
import sys

try:
    import pgdb
except ImportError:
    pgdb = None

try:
    import MySQLdb
except ImportError:
    MySQLdb = None

try:
    import sqlite3
except ImportError:
    sqlite3 = None

try:
    import pymongo
except ImportError:
    pymongo = None

from tdda.constraints.base import UNICODE_TYPE
from tdda.constraints.flags import (discover_parser, discover_flags,
                                    verify_parser, verify_flags)


DATABASE_USAGE = '''

Database connection flags:

  * -conn FILE              Database connection file
  * -dbtype DBTYPE          Type of database
  * -db DATABASE            Name of database to connect to
  * -host HOSTNAME          Name of server to connect to
  * -port PORTNUMBER        IP port number to connect to
  * -user USERNAME          Username to connect as
  * -password PASSWORD      Password to authenticate with

If -conn is provided, then none of the other options are required, and
the database connection details are read from the specified file.

If the database type is specified (with the -dbtype option, or by
prefixing the table name, such as postgresql:mytable), then a default
connection file .tdda_db_conn_DBTYPE (in your home directory) is used,
if present.

'''


def parse_table_name(table, dbtype):
    """
    split a qualified table name into its two parts: the database type
    and the table name.
    """
    if ':' in table:
        parts = table.split(':')
        if dbtype is None:
            dbtype = parts[0].lower()
        table = parts[1]
    return (table, dbtype)


def applicable(argv):
    """
    Does a command line include parameters that are applicable for databases
    that are supported by these extensions?
    """
    for i, a in enumerate(argv):
        if ':' in a:
            (table, dbtype) = parse_table_name(a, None)
            if dbtype in DATABASE_HANDLERS:
                return True
        elif a == '-dbtype':
            return (i < len(argv) - 1
                    and argv[i+1] in DATABASE_HANDLERS)
    return '-db' in argv


def database_arg_parser(create_parser, usage):
    parser = create_parser(usage + DATABASE_USAGE)
    parser.add_argument('-conn', '--conn', nargs=1,
                        help='database connection file')
    parser.add_argument('-dbtype', '--dbtype', nargs=1, help='database type')
    parser.add_argument('-db', '--db', nargs=1, help='database name')
    parser.add_argument('-host', '--host', nargs=1,
                        help='database server hostname')
    parser.add_argument('-port', '--port',
                        nargs=1, help='database server IP port')
    parser.add_argument('-user', '--user', nargs=1, help='username')
    parser.add_argument('-password', '--password', nargs=1, help='password')
    return parser


def database_arg_flags(create_flags, parser, args, params):
    params.update({
        'conn': None,
        'dbtype': None,
        'db': None,
        'host': None,
        'port': None,
        'user': None,
        'password': None,
    })
    flags = create_flags(parser, args, params)
    if flags.conn:
        params['conn'] = flags.conn[0]
    if flags.dbtype:
        params['dbtype'] = flags.dbtype[0]
    if flags.db:
        params['db'] = flags.db[0]
    if flags.host:
        params['host'] = flags.host[0]
    if flags.port:
        params['port'] = int(flags.port[0])
    if flags.user:
        params['user'] = flags.user[0]
    if flags.password:
        params['password'] = flags.password[0]
    return flags


def database_connection(table=None, conn=None, dbtype=None, db=None,
                        host=None, port=None, user=None, password=None):
    """
    Connect to a database, using an appropriate driver for the type
    of database specified.
    """
    if conn:
        defaults = Connection(conn)
    else:
        if dbtype:
            dbtypelower = dbtype.lower()
            connfile = os.path.join(os.path.expanduser('~'),
                                    '.tdda_db_conn' + '_' + dbtypelower)
            if dbtypelower == 'postgres' and not os.path.exists(connfile):
                connfile = os.path.join(os.path.expanduser('~'),
                                        '.tdda_db_conn_postgresql')
        else:
            connfile = os.path.join(os.path.expanduser('~'), '.tdda_db_conn')
        if os.path.exists(connfile):
            defaults = Connection(connfile)
        else:
            defaults = None
    if defaults:
        if dbtype is None:
            dbtype = defaults.dbtype
        if db is None:
            db = defaults.db
        if host is None:
            host = defaults.host
        if port is None:
            port = defaults.port
        if user is None:
            user = defaults.user
        if password is None:
            password = defaults.password

    if host and (':' in host) and port is None:
        parts = host.split(':')
        host = parts[0]
        port = int(parts[1])
    if port is not None:
        port = int(port)

    if dbtype is None:
        print('Database type required ("-dbtype type")', file=sys.stderr)
        sys.exit(1)
    if db is None:
        print('Database name required ("-db name")', file=sys.stderr)
        sys.exit(1)

    dbtypelower = dbtype.lower()
    if dbtypelower in DATABASE_CONNECTORS:
        connector = DATABASE_CONNECTORS[dbtypelower]
        return connector(host, port, db, user, password)
    else:
        print('Database type %s not supported' % dbtype, file=sys.stderr)
        sys.exit(1)


def database_connection_postgres(host, port, db, user, password):
    if pgdb:
        if port is not None:
            host = host + ':' + str(port)
        return pgdb.connect(host=host, database=db,
                            user=user, password=password)
    else:
        print('PostgreSQL driver not available', file=sys.stderr)
        sys.exit(1)


def database_connection_mysql(host, port, db, user, password):
    if MySQLdb:
        # TODO: should provide support for MySQL 'option-files' too.
        if host is None:
            host = 'localhost'
        if port is None:
            port = 3306
        if user is None:
            user = getpass.getuser()
        if password:
            try:
                return MySQLdb.connect(host=host, port=port, db=db,
                                       user=user, password=password)
            except:
                # some versions of the MySQL driver use different names
                return MySQLdb.connect(host=host, port=port, db=db,
                                       username=user, passwd=password)
        else:
            return MySQLdb.connect(host=host, port=port, db=db, user=user)
    else:
        print('MySQL driver not available', file=sys.stderr)


def database_connection_sqlite(host, port, db, user, password):
    if sqlite3:
        conn = sqlite3.connect(db)
        conn.create_function('regexp', 2, regex_matcher)
        return conn
    else:
        print('sqlite driver not available', file=sys.stderr)
        sys.exit(1)


def database_connection_mongodb(host, port, db, user, password):
    if pymongo:
        if host is None:
            host = 'localhost'
        if port is None:
            port = 27017
        client = pymongo.MongoClient(host, port)
        if db not in client.database_names():
            print('Mongodb database %s does not exist' % db)
            sys.exit(1)
        return client[db]
    else:
        print('mongodb driver not available', file=sys.stderr)
        sys.exit(1)


def regex_matcher(expr, item):
    """
    REGEXP implementation for Sqlite
    """
    if item is None:
        return False
    else:
        return re.match(expr, item) is not None


class Connection:
    """
    Class for reading a connection specification file.
    """
    def __init__(self, filename):
        with open(filename) as f:
            d = json.loads(f.read())
        self.dbtype = d.get('dbtype')
        self.db = d.get('db')
        self.host = d.get('host')
        self.port = d.get('port')
        self.user = d.get('user')
        self.password = d.get('password')

        if (self.dbtype and self.dbtype.lower() == 'sqlite'
                    and self.db is not None
                    and not os.path.isabs(self.db)):
            # if a .conn file for a sqlite connection specifies a .db file,
            # then resolve that relative to the location of the .conn file.
            self.db = os.path.join(os.path.dirname(filename), self.db)


class DatabaseHandler:
    """
    Common SQL and NoSQL database support
    """
    def __init__(self, dbtype, db):
        handlerClass = self.check_db_type(dbtype)
        self.instance = handlerClass(dbtype, db)

    def check_db_type(self, dbtype):
        """
        Check that this is one of the database types that we recognize
        """
        if dbtype in DATABASE_HANDLERS:
            return DATABASE_HANDLERS[dbtype]
        else:
            raise Exception('Unsupported database type')

    def __getattr__(self, name):
        return getattr(self.instance, name)


# TODO: this is too specific to Miro stuff
RESERVED_WORDS = ('group')


class SQLDatabaseHandler:
    """
    Common database SQL support
    """
    def __init__(self, dbtype, db):
        self.dbtype = dbtype
        self.db = db
        self.cursor = db.cursor()

    def canon(self, name):
        # 'canonicalise' a database table name
        name = name.lower()
        if name in RESERVED_WORDS:
            name = name + '_'
        return name

    def execute_scalar(self, sql):
        # execute a SQL statement, returning a single scalar result
        self.cursor.execute(sql)
        result = self.cursor.fetchall()[0][0]
        if result == '' and self.dbtype == 'sqlite':
            result = None
        return result

    def execute_all(self, sql):
        # execute a SQL statement, returning a list of rows
        self.cursor.execute(sql)
        return self.cursor.fetchall()

    def db_value_is_null(self, value):
        return value is None

    def db_value_to_datetime(self, value):
        return value

    def check_table_exists(self, tablename):
        """
        Check that a table (or a schema.table) exists.
        """
        if '.' in tablename:
            parts = tablename.split('.')
            if len(parts) == 1:
                schema = None
                table = tablename
            elif len(parts) == 2:
                schema = parts[0]
                table = parts[1]
            else:
                raise Exception('Bad table format %s' % tablename)
        else:
            schema = None
            table = tablename
        if self.dbtype in ('postgres', 'postgresql'):
            if schema is None:
                schema = 'public'
            sql = '''SELECT COUNT(*) FROM pg_catalog.pg_class c
                     LEFT JOIN pg_catalog.pg_namespace n
                     ON n.oid = c.relnamespace
                     WHERE c.relkind IN ('r', '', 'v')
                     AND c.relname = '%s'
                     AND n.nspname = '%s';''' % (table, schema)
        elif self.dbtype == 'mysql':
            if schema:
                sql = '''SELECT COUNT(*) FROM information_schema.tables
                         WHERE table_schema = '%s'
                         AND table_name = '%s';''' % (schema, table)
            else:
                sql = '''SELECT COUNT(*) FROM information_schema.tables
                         WHERE table_name = '%s';''' % table
        elif self.dbtype == 'sqlite':
            sql = '''SELECT COUNT(*) FROM sqlite_master
                            WHERE (type = 'table' OR type='view')
                            AND name = '%s';''' % tablename
        else:
            raise Exception('Unsupported database type %s' % self.dbtype)
        return self.execute_scalar(sql) > 0

    def get_database_column_names(self, tablename):
        if self.dbtype in ('postgres', 'postgresql'):
            sql = '''
                SELECT a.attname FROM pg_attribute a
                WHERE a.attrelid = '%s'::regclass AND a.attnum > 0
                ORDER BY a.attnum;
                ''' % tablename
            rows = self.execute_all(sql)
            return [r[0] for r in rows]
        elif self.dbtype == 'mysql':
            sql = '''
                SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS
                WHERE table_name = '%s';
                ''' % tablename
            rows = self.execute_all(sql)
            return [r[0] for r in rows]
        elif self.dbtype == 'sqlite':
            sql = 'PRAGMA table_info(%s)' % tablename
            rows = self.execute_all(sql)
            return [r[1] for r in rows]
        else:
            raise Exception('Unsupported database type')

    def get_database_column_type(self, tablename, colname):
        typeMap = {
            'int'        : 'int',
            'int4'       : 'int',
            'int8'       : 'int',
            'smallint'   : 'int',
            'bigint'     : 'int',
            'integer'    : 'int',
            'float'      : 'real',
            'float4'     : 'real',
            'float8'     : 'real',
            'float16'    : 'real',
            'double'     : 'real',
            'numeric'    : 'real',
            'real'       : 'real',
            'bool'       : 'bool',
            'text'       : 'string',
            'varchar'    : 'string',
            'char'       : 'string',
            'name'       : 'string',
            'oidvector'  : 'string',
            'timestamp'  : 'date',
            'date'       : 'date',
            None         : None,
        }
        if self.dbtype in ('postgres', 'postgresql'):
            sql = '''
                SELECT t.typname
                FROM pg_attribute a inner join pg_type t on a.atttypid = t.oid
                WHERE a.attrelid = '%s'::regclass AND a.attnum > 0
                                                  AND a.attname = '%s'
                ORDER BY a.attnum;
                ''' % (tablename, self.canon(colname))
            typeresult = self.execute_scalar(sql)
        elif self.dbtype == 'mysql':
            sql = '''
                SELECT DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS
                WHERE table_name = '%s' AND COLUMN_NAME = '%s';
                ''' % (tablename, self.canon(colname))
            typeresult = self.execute_scalar(sql)
        elif self.dbtype == 'sqlite':
            result = self.execute_all('PRAGMA table_info(%s)' % tablename)
            typeresult = None
            for row in result:
                if self.canon(row[1]) == self.canon(colname):
                    typeresult = row[2]
                    break
        else:
            raise Exception('Unsupported database type')
        return typeMap[typeresult.lower()]

    def get_database_nrows(self, tablename):
        sql = 'SELECT COUNT(*) FROM %s' % tablename
        return self.execute_scalar(sql)

    def get_database_nnull(self, tablename, colname):
        sql = ('SELECT COUNT(*) FROM %s WHERE %s IS NULL'
               % (tablename, self.canon(colname)))
        return self.execute_scalar(sql)

    def get_database_nnonnull(self, tablename, colname):
        sql = ('SELECT COUNT(*) FROM %s WHERE %s IS NOT NULL'
               % (tablename, self.canon(colname)))
        return self.execute_scalar(sql)

    def get_database_min(self, tablename, colname):
        sql = 'SELECT MIN(%s) FROM %s' % (self.canon(colname), tablename)
        return self.execute_scalar(sql)

    def get_database_max(self, tablename, colname):
        sql = 'SELECT MAX(%s) FROM %s' % (self.canon(colname), tablename)
        return self.execute_scalar(sql)

    def get_database_min_length(self, tablename, colname):
        sql = 'SELECT MIN(LENGTH(%s)) FROM %s' % (self.canon(colname),
                                                  tablename)
        return self.execute_scalar(sql)

    def get_database_max_length(self, tablename, colname):
        sql = 'SELECT MAX(LENGTH(%s)) FROM %s' % (self.canon(colname),
                                                  tablename)
        return self.execute_scalar(sql)

    def get_database_nunique(self, tablename, colname):
        colname = self.canon(colname)
        sql = ('SELECT COUNT(DISTINCT %s) FROM %s WHERE %s IS NOT NULL'
               % (colname, tablename, self.canon(colname)))
        return self.execute_scalar(sql)

    def get_database_unique_values(self, tablename, colname,
                                   sorted_values=True, include_nulls=False):
        colname = self.canon(colname)
        whereclause = ('' if include_nulls
                       else 'WHERE %s IS NOT NULL' % colname)
        orderby = ('ORDER BY %s ASC' % colname) if sorted_values else ''
        sql = 'SELECT DISTINCT %s FROM %s %s %s' % (colname, tablename,
                                                    whereclause, orderby)
        result = self.execute_all(sql)
        return [x[0] for x in result]

    def get_database_rex_match(self, tablename, colname, rexes):
        if rexes is None:      # a null value is not considered to be an
            return True        # active constraint, so is always satisfied
        name = self.canon(colname)
        if self.dbtype in ('postgres', 'postgresql'):
            # postgresql uses ~ syntax
            rexprs = ["(%s ~ '%s')" % (name, r) for r in rexes]
        elif self.dbtype == 'mysql':
            # mysql uses REGEXP syntax, and doesn't understand \d
            rexes = [r.replace('\\d', '[0-9]') for r in rexes]
            rexprs = ["(%s REGEXP '%s')" % (name, r) for r in rexes]
        elif self.dbtype == 'sqlite':
            # sqlite doesn't support regular expressions unless the
            # regexp() user-defined function is available - but we have
            # arranged for that in the database_connection_mongodb function.
            rexprs = ["(%s REGEXP '%s')" % (name, r) for r in rexes]
        else:
            raise Exception('Unsupported database type')

        sql = ('SELECT COUNT(*) FROM %s WHERE %s IS NOT NULL AND NOT(%s)'
               % (tablename, name, ' OR '.join(rexprs)))
        return self.execute_scalar(sql) == 0


class MongoDBDatabaseHandler:
    """
    NoSQL MonggoDB support
    """
    def __init__(self, dbtype, db):
        self.dbtype = dbtype
        self.db = db

    def find_collection(self, tablename):
        """
        Search through the collections hierarchy to resolve dotted names
        """
        parts = tablename.split('.')
        collection = self.db
        for p in parts:
            collection = collection[p]
        return collection

    def check_table_exists(self, tablename):
        try:
            self.find_collection(tablename)
            return True
        except pymongo.errors.InvalidName:
            return False

    def get_database_column_names(self, tablename):
        collection = self.find_collection(tablename)
        mapfn = '''function() {
                        var keys = [];
                        for (var k in this) {
                            keys.push(k);
                        }
                        emit(null, keys);
                   }'''
        redfn = '''function(key, values) {
                       var keyset = {};
                       for (var i = 0; i < values.length; i++) {
                           for (var j = 0; j < values[i].length; j++) {
                               keyset[values[i][j]] = true;
                           }
                       }
                       var keys = [];
                       for (var k in keyset) {
                           if (k != '_id') {
                               keys.push(k);
                           }
                       }
                       return JSON.stringify(keys);
                   }'''
        v = collection.map_reduce(mapfn, redfn, 'inline').find_one()['value']
        # this mongodb map/reduce op returns a json repr of the list of keys
        return json.loads(v)

    def get_database_column_type(self, tablename, colname):
        collection = self.find_collection(tablename)
        doc = collection.find_one({colname: {'$ne': None}})
        if doc:
            val = doc[colname]
            valtype = type(val)
            if valtype == bool:
                return 'bool'
            elif valtype in (int, long):
                return 'int'
            elif valtype == float:
                return 'real'
            elif valtype in (str, UNICODE_TYPE):
                return 'string'
            elif isinstance(val, datetime.datetime):
                return 'date'
            else:
                return 'other'
        else:
            return None

    def get_database_nrows(self, tablename):
        collection = self.find_collection(tablename)
        return collection.count()

    def get_database_nnull(self, tablename, colname):
        collection = self.find_collection(tablename)
        return collection.find({colname: {'$eq': None}}).count()

    def get_database_nnonnull(self, tablename, colname):
        collection = self.find_collection(tablename)
        return collection.find({colname: {'$ne': None}}).count()

    def get_database_nunique(self, tablename, colname):
        collection = self.find_collection(tablename)
        #return len(collection.distinct(colname))
        return collection.aggregate([
            {'$match': {colname: {'$exists': True}}},
            {'$match': {colname: {'$ne': None}}},
            {'$group': {'_id': '$' + colname}},
            {'$group': {'_id': None, 'nunique': {'$sum': 1}}},
        ]).next()['nunique']

    def get_database_unique_values(self, tablename, colname,
                                   sorted_values=True):
        collection = self.find_collection(tablename)
        values = collection.distinct(colname)
        return sorted(values) if sorted_values else values

    def get_database_min_length(self, tablename, colname):
        collection = self.find_collection(tablename)
        mapfn = 'function() { emit(null, this.name ? this.name.length : null);}'
        redfn = 'function(key, values) { return Math.min.apply(Math, values);}'
        v = collection.map_reduce(mapfn, redfn, 'inline').find_one()['value']
        return int(v)

    def get_database_max_length(self, tablename, colname):
        collection = self.find_collection(tablename)
        mapfn = 'function() { emit(null, this.name ? this.name.length: null);}'
        redfn = 'function(key, values) { return Math.max.apply(Math, values);}'
        v = collection.map_reduce(mapfn, redfn, 'inline').find_one()['value']
        return int(v)

    def get_database_min(self, tablename, colname):
        collection = self.find_collection(tablename)
        return collection.aggregate([
            {'$match': {colname: {'$exists': True}}},
            {'$group': {'_id': None, 'min': {'$min': '$' + colname}}},
        ]).next()['min']

    def get_database_max(self, tablename, colname):
        collection = self.find_collection(tablename)
        return collection.aggregate([
            {'$match': {colname: {'$exists': True}}},
            {'$group': {'_id': None, 'max': {'$max': '$' + colname}}},
        ]).next()['max']

    def db_value_is_null(self, value):
        return value is None

    def db_value_to_datetime(self, value):
        return value


DATABASE_CONNECTORS = {
    'postgres': database_connection_postgres,
    'postgresql': database_connection_postgres,
    'mysql': database_connection_mysql,
    'sqlite': database_connection_sqlite,
    'mongodb': database_connection_mongodb
}

DATABASE_HANDLERS = {
    'postgres': SQLDatabaseHandler,
    'postgresql': SQLDatabaseHandler,
    'mysql': SQLDatabaseHandler,
    'sqlite': SQLDatabaseHandler,
    'mongodb': MongoDBDatabaseHandler,
}




# elements_verify_118.py

from __future__ import print_function
import pandas as pd

from tdda.constraints.pd.constraints import verify_df

df = pd.read_csv('testdata/elements118.csv')
verification = verify_df(df, 'testdata/elements92.tdda')

print('Basic Verification:')
print(verification)
print('\n')
print('Verification DataFrame:')
dfv = verification.to_frame()
print(dfv)




# -*- coding: utf-8 -*-

"""
checkfiles.py: comparison mechanism for text files

Source repository: http://github.com/tdda/tdda

License: MIT

Copyright (c) Stochastic Solutions Limited 2016-2018
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
# from __future__ import unicode_literals

import os
import re
import sys
import tempfile


class BaseComparison(object):

    def __init__(self, print_fn=None, verbose=True, tmp_dir=None):
        """
        Constructor for an instance of the PandasComparison class.

        The optional print_fn parameter is a function to be used to
        display information while comparison operations are running.
        If specified, it should be a function with the same signature
        as python's builtin (__future__) print function.
        """
        self.print_fn = print_fn
        self.verbose = verbose
        self.tmp_dir = tmp_dir or tempfile.gettempdir()

    def info(self, msgs, s):
        """
        Add an item to the list of messages, and also display it immediately
        if verbose is set.
        """
        msgs.append(s)
        if self.verbose and self.print_fn:
            self.print_fn(s)

    @staticmethod
    def compare_with(actual, expected, qualifier=None):
        qualifier = '' if not qualifier else (qualifier + ' ')
        if os.path.exists(expected):
            msg = 'Compare %swith:\n    %s %s %s\n'
            cmd = diffcmd()
        else:
            msg = 'Initialize %sfrom actual content with:\n    %s %s %s'
            cmd = copycmd()
        return msg % (qualifier, cmd, actual, expected)


def diffcmd():
    return 'fc' if os.name and os.name != 'posix' else 'diff'


def copycmd():
    return 'copy' if os.name and os.name != 'posix' else 'cp'




# -*- coding: utf-8 -*-
"""
The :py:mod:`tdda.constraints.pd.constraints` module provides an
implementation of TDDA constraint discovery and verification
for Pandas Dataframes.

This allows it to be used for data in CSV files, or for Pandas or R
DataFrames saved as Feather files.

The top-level functions are:

    :py:func:`discover_df`:
        Discover constraints from a Pandas DataFrame.

    :py:func:`verify_df`:
        Verify (check) a Pandas DataFrame, against a set of previously
        discovered constraints.

    :py:func:`detect_df`:
        Verify (check) a Pandas DataFrame, against a set of previously
        discovered constraints, and generate an output dataset containing
        information about input rows which failed any of the constraints.

"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import datetime
import os
import re
import sys

from collections import OrderedDict

import pandas as pd
import numpy as np

try:
    from pmmif import featherpmm
except ImportError:
    featherpmm = None
    try:
        import feather as feather
    except ImportError:
        feather = None

from tdda.constraints.base import (
    STANDARD_FIELD_CONSTRAINTS,
    native_definite,
    DatasetConstraints,
    Verification,
    fuzz_up, fuzz_down,
)
from tdda.constraints.baseconstraints import (
    BaseConstraintCalculator,
    BaseConstraintDetector,
    BaseConstraintVerifier,
    BaseConstraintDiscoverer,
    MAX_CATEGORIES,
)

from tdda.referencetest.checkpandas import (default_csv_loader,
                                            default_csv_writer)
from tdda import rexpy

# pd.tslib is deprecated in newer versions of Pandas
if hasattr(pd, 'Timestamp'):
    pandas_Timestamp = pd.Timestamp
else:
    pandas_Timestamp = pd.tslib.Timestamp

isPy3 = sys.version_info[0] >= 3
if isPy3:
    long = int

DEBUG = False


class PandasConstraintCalculator(BaseConstraintCalculator):
    """
    Implementation of the Constraint Calculator methods for
    Pandas dataframes.
    """
    def __init__(self, df):
        self.df = df

    def is_null(self, value):
        return pd.isnull(value)

    def to_datetime(self, value):
        return pd.to_datetime(value)

    def get_column_names(self):
        return list(self.df)

    def get_nrecords(self):
        return len(self.df)

    def types_compatible(self, x, y, colname):
        return pandas_types_compatible(x, y, colname)

    def calc_min(self, colname):
        if self.df[colname].dtype == np.dtype('O'):
            m = self.df[colname].dropna().min()  # Otherwise -inf!
        else:
            m = self.df[colname].min()
        if pandas_tdda_type(m) == 'date':
            m = m.to_pydatetime()
        elif hasattr(m, 'item'):
            m = m.item()
        return m

    def calc_max(self, colname):
        if self.df[colname].dtype == np.dtype('O'):
            M = self.df[colname].dropna().max()
        else:
            M = self.df[colname].max()
        if pandas_tdda_type(M) == 'date':
            M = M.to_pydatetime()
        elif hasattr(M, 'item'):
            M = M.item()
        return M

    def calc_min_length(self, colname):
        if isPy3:
            return self.df[colname].str.len().min()
        else:
            return self.df[colname].str.decode('UTF-8').str.len().min()

    def calc_max_length(self, colname):
        if isPy3:
            return self.df[colname].str.len().max()
        else:
            return self.df[colname].str.decode('UTF-8').str.len().max()

    def calc_tdda_type(self, colname):
        return pandas_tdda_column_type(self.df[colname])

    def calc_null_count(self, colname):
        return int(len(self.df) - self.df[colname].count())

    def calc_non_null_count(self, colname):
        return int(len(self.df) - self.calc_null_count(colname))

    def calc_nunique(self, colname):
        return int(self.df[colname].nunique())

    def calc_unique_values(self, colname, include_nulls=True):
        values = self.df[colname].unique()
        nullvalues = ([v for v in self.df[colname].unique()
                         if pd.isnull(v)] if include_nulls
                      else [])
        nonnullvalues = [v for v in values if not pd.isnull(v)]
        return nullvalues + sorted(nonnullvalues)

    def calc_non_integer_values_count(self, colname):
        values = self.df[colname].dropna()
        non_nulls = self.df[colname].count()
        return int(non_nulls
                   - (values.astype(int) == values).astype(int).sum())

    def calc_all_non_nulls_boolean(self, colname):
        nn = self.df[colname].dropna()
        return all([type(v) is bool for i, v in nn.iteritems()])

    def allowed_values_exclusions(self):
        # remarkably, Pandas returns various kinds of nulls as
        # unique values, despite not counting them with .nunique()
        return [None, np.nan, pd.NaT]

    def find_rexes(self, colname, values=None):
        if values is None:
            return rexpy.pdextract(self.df[colname])
        else:
            return rexpy.extract(values)

    def calc_rex_constraint(self, colname, constraint, detect=False):
        # note that this should return a set of violations, not True/False.
        rexes = constraint.value
        if rexes is None:      # a null value is not considered
            return None        # to be an active constraint,
                               # so is always satisfied
        rexes = [re.compile(r) for r in rexes]
        strings = [native_definite(s)
                   for s in self.df[colname].dropna().unique()]

        failures = set()
        for s in strings:
            for r in rexes:
                if re.match(r, s):
                    break
            else:
                if DEBUG:
                    print('*** Unmatched string: "%s"' % s)
                if detect:
                    failures.add(s)
                else:
                    return True  # At least one string didn't match
        if detect:
            return failures
        else:
            return None


class PandasConstraintDetector(BaseConstraintDetector):
    """
    Implementation of the Constraint Detector methods for
    Pandas dataframes.
    """
    def __init__(self, df):
        self.df = df
        self.out_df = (pd.DataFrame(index=df.index.copy()) if df is not None
                       else None)

    def detect_min_constraint(self, col, constraint, precision, epsilon):
        name = unique_column_name(self.df, col + '_min_ok')
        value = constraint.value
        c = self.df[col]
        if precision == 'closed':
            self.out_df[name] = detection_field(c, c >= value)
        elif precision == 'open':
            self.out_df[name] = detection_field(c, c > value)
        else:
            self.out_df[name] = detection_field(c, df_fuzzy_gt(c, value,
                                                               epsilon))
            print(self.out_df[name].dtype)

    def detect_max_constraint(self, col, constraint, precision, epsilon):
        name = unique_column_name(self.df, col + '_max_ok')
        value = constraint.value
        c = self.df[col]
        if precision == 'closed':
            self.out_df[name] = detection_field(c, c <= value)
        elif precision == 'open':
            self.out_df[name] = detection_field(c, c < value)
        else:
            self.out_df[name] = detection_field(c, df_fuzzy_lt(c, value,
                                                               epsilon))

    def detect_min_length_constraint(self, col, constraint):
        name = unique_column_name(self.df, col + '_min_length_ok')
        value = constraint.value
        c = self.df[col]
        self.out_df[name] = detection_field(c, c.str.len() >= value)

    def detect_max_length_constraint(self, col, constraint):
        name = unique_column_name(self.df, col + '_max_length_ok')
        value = constraint.value
        c = self.df[col]
        self.out_df[name] = detection_field(c, c.str.len() <= value)

    def detect_tdda_type_constraint(self, col, constraint):
        name = unique_column_name(self.df, col + '_type_ok')
        self.out_df[name] = detection_field(None, False)

    def detect_sign_constraint(self, col, constraint):
        name = unique_column_name(self.df, col + '_sign_ok')
        value = constraint.value
        c = self.df[col]

        if value == 'null':
            self.out_df[name] = detection_field(None, False)
        elif value == 'positive':
            self.out_df[name] = detection_field(c, c > 0)
        elif value == 'non-negative':
            self.out_df[name] = detection_field(c, c >= 0)
        elif value == 'zero':
            self.out_df[name] = detection_field(c, c == 0)
        elif value == 'non-positive':
            self.out_df[name] = detection_field(c, c <= 0)
        elif value == 'negative':
            self.out_df[name] = detection_field(c, c < 0)

    def detect_max_nulls_constraint(self, col, constraint):
        # found more nulls than are allowed, so mark all null values as bad
        name = unique_column_name(self.df, col + '_nonnull_ok')
        c = self.df[col]
        self.out_df[name] = pd.notnull(c)

    def detect_no_duplicates_constraint(self, col, constraint):
        # found duplicates, so mark anything duplicated as bad
        name = unique_column_name(self.df, col + '_nodups_ok')
        c = self.df[col]
        unique = ~ self.df.duplicated(col, keep=False)
        self.out_df[name] = detection_field(c, unique, default=True)

    def detect_allowed_values_constraint(self, col, constraint, violations):
        name = unique_column_name(self.df, col + '_values_ok')
        c = self.df[col]
        self.out_df[name] = detection_field(c, ~ c.isin(violations))

    def detect_rex_constraint(self, col, constraint, violations):
        name = unique_column_name(self.df, col + '_rex_ok')
        c = self.df[col]
        self.out_df[name] = detection_field(c, ~ c.isin(violations))

    def write_detected_records(self,
                               detect_outpath=None,
                               detect_write_all=False,
                               detect_per_constraint=False,
                               detect_output_fields=None,
                               detect_rownumber=False,
                               detect_in_place=False,
                               **kwargs):
        if self.out_df is None:
            return None

        out_df = self.out_df
        add_index = detect_rownumber or detect_output_fields is None
        if detect_output_fields is None:
            detect_output_fields = []
        elif len(detect_output_fields) == 0:
            detect_output_fields = list(self.df)

        nfailname = unique_column_name(self.df, 'n_failures')
        nf = len(list(out_df))
        fails = (nf - out_df.sum(axis=1).astype(float)
                    - out_df.isnull().sum(axis=1).astype(float))
        out_df[nfailname] = fails.astype(int)

        if detect_in_place:
            for fname in list(out_df):
                self.df[unique_column_name(self.df, fname)] = out_df[fname]
        if not detect_write_all:
            out_df = out_df[out_df.n_failures > 0]
        if not detect_per_constraint:
            fnames = [name for name in list(out_df) if name != nfailname]
            out_df = out_df.drop(fnames, axis=1)
        if detect_output_fields:
            for fname in reversed(detect_output_fields):
                if fname in list(self.df):
                    out_df.insert(0, fname, self.df[fname])
                else:
                    raise Exception('Dataframe has no column %s' % fname)

        if detect_outpath:
            if add_index:
                rownumbername = unique_column_name(out_df, 'RowNumber')
                out_df.insert(0, rownumbername, pd.RangeIndex(1, len(out_df)+1))
            save_df(out_df, detect_outpath)

        return out_df


class PandasConstraintVerifier(PandasConstraintCalculator,
                               PandasConstraintDetector,
                               BaseConstraintVerifier):
    """
    A :py:class:`PandasConstraintVerifier` object provides methods
    for verifying every type of constraint against a Pandas DataFrame.
    """
    def __init__(self, df, epsilon=None, type_checking=None):
        PandasConstraintCalculator.__init__(self, df)
        PandasConstraintDetector.__init__(self, df)
        BaseConstraintVerifier.__init__(self, epsilon=epsilon,
                                        type_checking=type_checking)

    def repair_field_types(self, constraints):
        # We sometimes haven't inferred the field types correctly for
        # the dataframe (e.g. if we read it from a csv file, "string"
        # fields might look like numeric ones, if they only contain digits).
        # We can try to use the constraint information to try to repair this,
        # but it's not always going to be successful.
        for c in self.df.columns.tolist():
            if c not in constraints:
                continue
            ser = self.df[c]
            try:
                ctype = constraints[c]['type'].value
                dtype = ser.dtype
                if ctype == 'string' and dtype != pd.np.dtype('O'):
                    is_numeric = True
                    is_real = False
                    for limit in ('min', 'max'):
                        if limit in constraints[c]:
                            limitval = constraints[c][limit].value
                            if type(limitval) in (int, long, float):
                                if type(limitval) == float:
                                    is_real = True
                            else:
                                is_numeric = False
                                break
                    if is_numeric:
                        if is_real:
                            is_real = self.calc_non_integer_values_count(c) > 0
                        self.df.loc[ser.notnull(), c] = ser.astype(str)
                        if not is_real:
                            self.df[c] = self.df[c].str.replace('.0', '')
                elif ctype == 'bool' and dtype == pd.np.dtype('int64'):
                    self.df[c] = ser.astype(bool)
                elif ctype == 'bool' and dtype == pd.np.dtype('int32'):
                    self.df[c] = ser.astype(bool)
            except Exception as e:
                print('%s: %s' % (e.__class__.__name__, str(e)))
                pass


class PandasVerification(Verification):
    """
    A :py:class:`PandasVerification` object adds a :py:meth:`to_frame()`
    method to a :py:class:`tdda.constraints.base.Verification` object.

    This allows the result of constraint verification to be converted to a
    Pandas DataFrame, including columns for the field (column) name,
    the numbers of passes and failures and boolean columns for each
    constraint, with values:

    - ``True``       --- if the constraint was satified for the column
    - ``False``      --- if column failed to satisfy the constraint
    - ``pd.np.NaN``  --- if there was no constraint of this kind

    This Pandas-specific implementation of constraint verification also
    provides methods :py:meth:`to_frame` to get the overall verification
    result as as a Pandas DataFrame, and :py:meth:`detected` to get any
    detection results as a a Pandas DataFrame (if the verification has been
    run with in ``detect`` mode).
    """
    def __init__(self, *args, **kwargs):
        Verification.__init__(self, *args, **kwargs)

    def to_frame(self):
        """
        Converts object to a Pandas DataFrame.
        """
        return self.verification_to_dataframe(self)

    @staticmethod
    def verification_to_dataframe(ver):
        fields = ver.fields
        df = pd.DataFrame(OrderedDict((
            ('field', list(fields.keys())),
            ('failures', [v.failures for k, v in fields.items()]),
            ('passes', [v.passes for k, v in fields.items()]),
        )))
        kinds_used = set([])
        for field, constraints in fields.items():
            kinds_used = kinds_used.union(set(list(constraints.keys())))
        base_kinds = [k for k in STANDARD_FIELD_CONSTRAINTS if k in kinds_used]
        other_kinds = [k for k in kinds_used if not k in base_kinds]
        for kind in base_kinds + other_kinds:
            df[kind] = [fields[field].get(kind, np.nan) for field in fields]
        return df

    to_dataframe = to_frame


class PandasDetection(PandasVerification):
    """
    A :py:class:`PandasDetection` object adds a :py:meth:`detected()`
    method to a :py:class:`PandasVerification` object.

    This allows the Pandas DataFrame resulting from constraint detection
    to be made available.
    """
    def detected(self):
        """
        Returns the Pandas DataFrame containing the detection column,
        if the verification process has been run with in ``detect`` mode.
        """
        return getattr(self, 'detection', None)


class PandasConstraintDiscoverer(PandasConstraintCalculator,
                                 BaseConstraintDiscoverer):
    """
    A :py:class:`PandasConstraintDiscoverer` object is used to discover
    constraints on a Pandas DataFrame.
    """
    def __init__(self, df, inc_rex=False):
        PandasConstraintCalculator.__init__(self, df)
        BaseConstraintDiscoverer.__init__(self, inc_rex=inc_rex)


def pandas_tdda_column_type(x):
    """
    Returns the TDDA type of a column.

    Basic TDDA types are one of 'bool', 'int', 'real', 'string' or 'date'.

    If *x* is ``None`` or something Pandas classes as null, 'null' is returned.

    If *x* is not recognized as one of these, 'other' is returned.
    """
    dt = getattr(x, 'dtype', None)
    if dt == np.dtype('O'):
        return 'string'
    dts = str(dt)
    if 'bool' in dts:
        return 'bool'
    if 'int' in dts:
        return 'int'
    if 'float' in dts:
        return 'real'
    if 'datetime' in dts:
        return 'date'
    if not isinstance(x, pd.core.series.Series) and pd.isnull(x):
        return 'null'
    # Everything else is other, for now, including compound types,
    # unicode in Python2, bytes in Python3 etc.
    return 'other'


def pandas_types_compatible(x, y, colname=None):
    """
    Returns boolean indicating whether the coarse_type of *x* and *y* are
    the same, for scalar values.

    If *colname* is provided, and the check fails, a warning is issued
    to stderr.
    """
    ok = pandas_coarse_type(x) == pandas_coarse_type(y)
    if not ok and colname:
        print('Warning: Failing incompatible types constraint for field %s '
              'of type %s.\n(Constraint value %s of type %s.)'
              % (colname, type(x), y, type(y)), file=sys.stderr)
    return ok


def pandas_coarse_type(x):
    """
    Returns the TDDA coarse type of *x*, a scalar value.
    The coarse types combine ``bool``, ``int`` and ``real`` into ``number``.

    Obviously, some people will dislike treating booleans as numbers.
    But it is necessary here.
    """
    t = pandas_tdda_type(x)
    return 'number' if t in ('bool', 'int', 'real') else t


def pandas_tdda_type(x):
    dt = getattr(x, 'dtype', None)
    if type(x) == str or dt == np.dtype('O'):
        return 'string'
    dts = str(dt)
    if type(x) == bool or 'bool' in dts:
        return 'bool'
    if type(x) in (int, long) or 'int' in dts:
        return 'int'
    if type(x) == float or 'float' in dts:
        return 'real'
    if (type(x) == datetime.datetime or 'datetime' in dts
                or type(x) == pandas_Timestamp):
        return 'date'
    if x is None:
        return 'null'
    null = pd.isnull(x)
    if hasattr(null, 'size'):
        null = False  # pd.isnull returned an array
    if (not isinstance(x, pd.core.series.Series) and null):
        return 'null'
    # Everything else is other, for now, including compound types,
    # unicode in Python2, bytes in Python3 etc.
    return 'other'


def verify_df(df, constraints_path, epsilon=None, type_checking=None,
              report='all', **kwargs):
    """
    Verify that (i.e. check whether) the Pandas DataFrame provided
    satisfies the constraints in the JSON ``.tdda`` file provided.

    Mandatory Inputs:

        *df*:
                            A Pandas DataFrame, to be checked.

        *constraints_path*:
                            The path to a JSON ``.tdda`` file (possibly
                            generated by the discover_df function, below)
                            containing constraints to be checked.

    Optional Inputs:

        *epsilon*:
                            When checking minimum and maximum values
                            for numeric fields, this provides a
                            tolerance. The tolerance is a proportion
                            of the constraint value by which the
                            constraint can be exceeded without causing
                            a constraint violation to be issued.

                            For example, with epsilon set to 0.01 (i.e. 1%),
                            values can be up to 1% larger than a max constraint
                            without generating constraint failure,
                            and minimum values can be up to 1% smaller
                            that the minimum constraint value without
                            generating a constraint failure. (These
                            are modified, as appropriate, for negative
                            values.)

                            If not specified, an *epsilon* of 0 is used,
                            so there is no tolerance.


                            NOTE: A consequence of the fact that these
                            are proportionate is that min/max values
                            of zero do not have any tolerance, i.e.
                            the wrong sign always generates a failure.

        *type_checking*:
                            ``strict`` or ``sloppy``.
                            Because Pandas silently, routinely and
                            automatically "promotes" integer and boolean
                            columns to reals and objects respectively
                            if they contain nulls, strict type checking
                            can be problematical in Pandas. For this reason,
                            ``type_checking`` defaults to ``sloppy``, meaning
                            that type changes that could plausibly be
                            attributed to Pandas type promotion will not
                            generate constraint values.

                            If this is set to strict, a Pandas ``float``
                            column ``c`` will only be allowed to satisfy a
                            an ``int`` type constraint if::

                                c.dropnulls().astype(int) == c.dropnulls()

                            Similarly, Object fields will satisfy a
                            ``bool`` constraint only if::

                                c.dropnulls().astype(bool) == c.dropnulls()

        *report*:
                            ``all`` or ``fields``.
                            This controls the behaviour of the
                            :py:meth:`~PandasVerification.__str__` method on
                            the resulting :py:class:`~PandasVerification`
                            object (but not its content).

                            The default is ``all``, which means that
                            all fields are shown, together with the
                            verification status of each constraint
                            for that field.

                            If report is set to ``fields``, only fields for
                            which at least one constraint failed are shown.

                            NOTE: The method also accepts two further
                            parameters to control (not yet implemented)
                            behaviour. ``constraints``, will be used to
                            indicate that only failing constraints for
                            failing fields should be shown.
                            ``one_per_line`` will indicate that each constraint
                            failure should be reported on a separate line.

    The various *detect* parameters from :py:meth:`detect_df` can also be
    used, in which case the verification process will also generate
    detection results.

    Returns:

        :py:class:`~PandasVerification` object.

        This object has attributes:

        - *passed*      --- Number of passing constriants
        - *failures*    --- Number of failing constraints

        It also has a :py:meth:`~PandasVerification.to_frame()` method for
        converting the results of the verification to a Pandas DataFrame,
        and a :py:meth:`~PandasVerification.__str__` method to print
        both the detailed and summary results of the verification.

    Example usage::

        import pandas as pd
        from tdda.constraints import verify_df

        df = pd.DataFrame({'a': [0, 1, 2, 10, pd.np.NaN],
                           'b': ['one', 'one', 'two', 'three', pd.np.NaN]})
        v = verify_df(df, 'example_constraints.tdda')

        print('Constraints passing: %d\\n' % v.passes)
        print('Constraints failing: %d\\n' % v.failures)
        print(str(v))
        print(v.to_frame())

    See *simple_verification.py* in the :ref:`constraint_examples`
    for a slightly fuller example.

    """
    pdv = PandasConstraintVerifier(df, epsilon=epsilon,
                                   type_checking=type_checking)
    constraints = DatasetConstraints(loadpath=constraints_path)
    pdv.repair_field_types(constraints)
    return pdv.verify(constraints,
                      VerificationClass=PandasVerification, **kwargs)


def detect_df(df, constraints_path, epsilon=None, type_checking=None,
              outpath=None, write_all=False, per_constraint=False,
              output_fields=None, rownumber=False, in_place=False, **kwargs):
    """
    Verify that (i.e. check whether) the Pandas DataFrame provided
    satisfies the constraints in the JSON ``.tdda`` file provided.

    Mandatory Inputs:

        *df*:
                            A Pandas DataFrame, to be checked.

        *constraints_path*:
                            The path to a JSON ``.tdda`` file (possibly
                            generated by the discover_df function, below)
                            containing constraints to be checked.

    Optional Inputs:

        *epsilon*:
                            When checking minimum and maximum values
                            for numeric fields, this provides a
                            tolerance. The tolerance is a proportion
                            of the constraint value by which the
                            constraint can be exceeded without causing
                            a constraint violation to be issued.

                            For example, with epsilon set to 0.01 (i.e. 1%),
                            values can be up to 1% larger than a max constraint
                            without generating constraint failure,
                            and minimum values can be up to 1% smaller
                            that the minimum constraint value without
                            generating a constraint failure. (These
                            are modified, as appropriate, for negative
                            values.)

                            If not specified, an *epsilon* of 0 is used,
                            so there is no tolerance.


                            NOTE: A consequence of the fact that these
                            are proportionate is that min/max values
                            of zero do not have any tolerance, i.e.
                            the wrong sign always generates a failure.

        *type_checking*:
                            ``strict`` or ``sloppy``.
                            Because Pandas silently, routinely and
                            automatically "promotes" integer and boolean
                            columns to reals and objects respectively
                            if they contain nulls, strict type checking
                            can be problematical in Pandas. For this reason,
                            ``type_checking`` defaults to ``sloppy``, meaning
                            that type changes that could plausibly be
                            attributed to Pandas type promotion will not
                            generate constraint values.

                            If this is set to strict, a Pandas ``float``
                            column ``c`` will only be allowed to satisfy a
                            an ``int`` type constraint if::

                                c.dropnulls().astype(int) == c.dropnulls()

                            Similarly, Object fields will satisfy a
                            ``bool`` constraint only if::

                                c.dropnulls().astype(bool) == c.dropnulls()

        *outpath*:
                            This specifies that the verification process
                            should detect records that violate any constraints,
                            and write them out to this CSV (or feather) file.

                            By default, only failing records are written out
                            to file, but this can be overridden with the
                            ``write_all`` parameter.

                            By default, the columns in the detection output
                            file will be a boolean ``ok`` field for each
                            constraint on each field, an and ``n_failures``
                            field containing the total number of constraints
                            that failed for each row.  This behavious can be
                            overridden with the ``per_constraint``,
                            ``output_fields`` and ``rownumber`` parameters.

        *write_all*:
                            Include passing records in the detection output
                            file when detecting.

        *per_constraint*:
                            Write one column per failing constraint, as well
                            as the ``n_failures`` total.

        *output_fields*:
                            Specify original columns to write out when detecting.

                            If passed in as an empty list (rather than None),
                            all original columns will be included.

        *rownumber*:
                            Include a row-number in the output file when
                            detecting.

                            This is automatically enabled if no output field
                            names are specified.

                            Rows are numbered from 0.

        *in_place*:
                            Detect failing constraints by adding columns to
                            the input DataFrame.

                            If ``outpath`` is also specified, then
                            failing constraints will also be written to file.

    The *report* parameter from :py:meth:`verify_df` can also be
    used, in which case a verification report will also be produced in
    addition to the detection results.

    Returns:

        :py:class:`~PandasDetection` object.

        This object has a :py:meth:`~PandasDetection.detected()` method
        for obtaining the Pandas Dataframe containing the detection
        results.

    Example usage::

        import pandas as pd
        from tdda.constraints import detect_df

        df = pd.DataFrame({'a': [0, 1, 2, 10, pd.np.NaN],
                           'b': ['one', 'one', 'two', 'three', pd.np.NaN]})
        v = detect_df(df, 'example_constraints.tdda')
        detection_df = v.detected()
        print(detection_df.to_string())

    """
    pdv = PandasConstraintVerifier(df, epsilon=epsilon,
                                   type_checking=type_checking)
    constraints = DatasetConstraints(loadpath=constraints_path)
    pdv.repair_field_types(constraints)
    return pdv.detect(constraints, VerificationClass=PandasDetection,
                      outpath=outpath, write_all=write_all,
                      per_constraint=per_constraint,
                      output_fields=output_fields, rownumber=rownumber,
                      in_place=in_place, **kwargs)


def discover_df(df, inc_rex=False):
    """
    Automatically discover potentially useful constraints that characterize
    the Pandas DataFrame provided.

    Input:

        *df*:
            any Pandas DataFrame.

        *inc_rex*:
            If ``True``, include discovery of regular expressions
            for string fields, using rexpy (default: ``False``).

    Possible return values:

    -  :py:class:`~tdda.constraints.base.DatasetConstraints` object
    -  ``None``    --- (if no constraints were found).

    This function goes through each column in the DataFrame and, where
    appropriate, generates constraints that describe (and are satisified
    by) this dataframe.

    Assuming it generates at least one constraint for at least one field
    it returns a :py:class:`tdda.constraints.base.DatasetConstraints` object.

    This includes a ``fields`` attribute, keyed on the column name.

    The returned :py:class:`~tdda.constraints.base.DatasetConstraints` object
    includes a :py:meth:`~tdda.constraints.base.DatasetContraints.to_json`
    method, which converts the constraints into JSON for saving as a tdda
    constraints file. By convention, such JSON files use a ``.tdda``
    extension.

    The JSON constraints file can be used to check whether other datasets
    also satisfy the constraints.

    The kinds of constraints (potentially) generated for each field (column)
    are:

        ``type``:
                the (coarse, TDDA) type of the field. One of
                ``bool``, ``int``, ``real``, ``string`` or ``date``.


        ``min``:
                for non-string fields, the minimum value in the column.
                Not generated for all-null columns.

        ``max``:
                for non-string fields, the maximum value in the column.
                Not generated for all-null columns.

        ``min_length``:
                For string fields, the length of the shortest string(s)
                in the field. N.B. In Python2, this assumes the strings
                are encoded in UTF-8, and an error may occur if this is
                not the case. String length counts unicode characters,
                not bytes.

        ``max_length``:
                For string fields, the length of the longest string(s)
                in the field.  N.B. In Python2, this assumes the strings
                are encoded in UTF-8, and an error may occur if this is
                not the case. String length counts unicode characters,
                not bytes.

        ``sign``:
                If all the values in a numeric field have consistent sign,
                a sign constraint will be written with a value chosen from:

                - ``positive``     --- For all values ``v`` in field: ``v > 0``
                - ``non-negative`` --- For all values ``v`` in field: ``v >= 0``
                - ``zero``         --- For all values ``v`` in field: ``v == 0``
                - ``non-positive`` --- For all values ``v`` in field: ``v <= 0``
                - ``negative``     --- For all values ``v`` in field: ``v < 0``
                - ``null``         --- For all values ``v`` in field: ``v is null``

        ``max_nulls``:
                The maximum number of nulls allowed in the field.

                - If the field has no nulls, a constraint
                  will be written with ``max_nulls`` set to zero.
                - If the field has a single null, a constraint will
                  be written with ``max_nulls`` set to one.
                - If the field has more than 1 null, no constraint
                  will be generated.

        ``no_duplicates``:
                For string fields (only, for now), if every
                non-null value in the field is different,
                this constraint will be generated (with value ``True``);
                otherwise no constraint will be generated. So this constraint
                indicates that all the **non-null** values in a string
                field are distinct (unique).

        ``allowed_values``:
                 For string fields only, if there are
                 :py:const:`MAX_CATEGORIES` or fewer distinct string
                 values in the dataframe, an AllowedValues constraint
                 listing them will be generated.
                 :py:const:`MAX_CATEGORIES` is currently "hard-wired" to 20.

    Example usage::

        import pandas as pd
        from tdda.constraints import discover_df

        df = pd.DataFrame({'a': [1, 2, 3], 'b': ['one', 'two', pd.np.NaN]})
        constraints = discover_df(df)
        with open('example_constraints.tdda', 'w') as f:
            f.write(constraints.to_json())

    See *simple_generation.py* in the :ref:`constraint_examples`
    for a slightly fuller example.

    """
    disco = PandasConstraintDiscoverer(df, inc_rex=inc_rex)
    return disco.discover()


def load_df(path):
    if os.path.splitext(path)[1] != '.feather':
        return default_csv_loader(path)
    elif featherpmm:
        ds = featherpmm.read_dataframe(path)
        return ds.df
    elif feather:
        return feather.read_dataframe(path)
    else:
        raise Exception('The Python feather module is not installed.\n'
                        'Use:\n    pip install feather-format\n'
                        'to add capability.\n', file=sys.stderr)


def save_df(df, path):
    if os.path.splitext(path)[1] != '.feather':
        default_csv_writer(df, path)
    elif featherpmm:
        featherpmm.write_dataframe(featherpmm.Dataset(df, name='verification'),
                                   path)
    elif feather:
        feather.write_dataframe(df, path)
    else:
        raise Exception('The Python feather module is not installed.\n'
                        'Use:\n    pip install feather-format\n'
                        'to add capability.\n', file=sys.stderr)


def unique_column_name(df, name):
    """
    Generate a column name that is not already present in the dataframe.
    """
    i = 1
    newname = name
    while newname in list(df):
        i += 1
        newname = '%s_%d' % (name, i)
    return newname


def detection_field(column, expr, default=None):
    """
    Construct a field for a detection result
    """
    default_value = (np.nan if default is None
                            else (np.ones(len(column)) * default))
    return np.where(pd.isnull(column), default_value, expr.astype('O'))


def df_fuzzy_gt(a, b, epsilon):
    """
    Returns a >~ b (a is greater than or approximately equal to b)

    At the moment, this simply reduces b by 1% if it is positive,
    and makes it 1% more negative if it is negative.
    """
    return (a >= b) | (a >= fuzz_down(b, epsilon))


def df_fuzzy_lt(a, b, epsilon):
    """
    Returns a <~ b (a is less than or approximately equal to b)

    At the moment, this increases b by 1% if it is positive,
    and makes it 1% less negative if it is negative.
    """
    return (a <= b) | (a <= fuzz_up(b, epsilon))


# for backwards compatibility (old name for function)
discover_constraints = discover_df


def applicable(argv):
    """
    Does a command line include parameters that are applicable for TDDA
    constraints for Pandas DataFrames or CSV files?
    """
    for i, a in enumerate(argv):
        if a.endswith('.csv') or a.endswith('.feather'):
            return True
    return False





import unittest

def common_string_sequence(s, t):
    """
    Find a sequence of characters that occurs, in order, in both s and t.
    The sequence need not be contiguous in either.

    Doesn't necessarily return the "best" (or longest) such sequence.
    """
    common = []
    p = 0         # position in string t
    for c in s:
        pos = t[p:].find(c)
        if pos > -1:
            common.append(c)
            p += pos + 1
    return ''.join(common)


def common_list_sequence(s, t):
    """
    Find a sequence of elements that occurs, in order, in both lists s and t.
    The sequence need not be contiguous in either.
    """
    common = []
    p = 0         # position in list t
    for c in s:
        pos = find(t[p:], c)
        if pos > -1:
            common.append(c)
            p += pos + 1
    return common


def find(L, c):
    return L.index(c) if c in L else -1


class TestCommonSequences(unittest.TestCase):
    def test_strings(self):
        self.assertEqual(common_string_sequence('', ''), '')
        self.assertEqual(common_string_sequence('abc', 'abc'), 'abc')
        self.assertEqual(common_string_sequence('abcde', 'ace'), 'ace')
        self.assertEqual(common_string_sequence('ace', 'abcde'), 'ace')
        self.assertEqual(common_string_sequence('abcde', 'aced'), 'acd')
        self.assertEqual(common_string_sequence('aced', 'abcde'), 'ace')
        self.assertEqual(common_string_sequence('aced', 'xyz'), '')
        self.assertEqual(common_string_sequence('aaaa', 'a'), 'a')
        self.assertEqual(common_string_sequence('a', 'aaaa'), 'a')
        self.assertEqual(common_string_sequence('aaaa', 'aa'), 'aa')
        self.assertEqual(common_string_sequence('aa', 'aaaa'), 'aa')
        self.assertEqual(common_string_sequence('abc', ''), '')
        self.assertEqual(common_string_sequence('', 'abc'), '')

    def test_sequences(self):
        a = 1
        b = 2
        c = 3
        d = 4
        e = 5
        x = 6
        y = 7
        z = 8
        self.assertEqual(common_list_sequence([], []), [])
        self.assertEqual(common_list_sequence([a, b, c], [a, b, c]), [a, b, c])
        self.assertEqual(common_list_sequence([a, b, c, d, e], [a, c, e]),
                                              [a, c, e])
        self.assertEqual(common_list_sequence([a, c, e], [a, b, c, d, e]),
                                              [a, c, e])
        self.assertEqual(common_list_sequence([a, b, c, d, e], [a, c, e, d]),
                                              [a, c, d])
        self.assertEqual(common_list_sequence([a, c, e, d], [a, b, c, d, e]),
                                              [a, c, e])
        self.assertEqual(common_list_sequence([a, c, e, d], [x, y, z]), [])
        self.assertEqual(common_list_sequence([a, a, a, a], [a]), [a])
        self.assertEqual(common_list_sequence([a], [a, a, a, a]), [a])
        self.assertEqual(common_list_sequence([a, a, a, a], [a, a]), [a, a])
        self.assertEqual(common_list_sequence([a, a], [a, a, a, a]), [a, a])
        self.assertEqual(common_list_sequence([a, b, c], []), [])
        self.assertEqual(common_list_sequence([], [a, b, c]), [])

if __name__ == '__main__':
     unittest.main()



# -*- coding: utf-8 -*-

#
# Unit tests for string functions from tdda.referencetest.checkfiles
#

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import division

import unittest

from tdda.referencetest.checkfiles import FilesComparison


class TestStrings(unittest.TestCase):

    def test_strings_ok(self):
        compare = FilesComparison()
        self.assertEqual(compare.check_strings([], []), (0, []))
        self.assertEqual(compare.check_strings(['abc'], ['abc']), (0, []))
        self.assertEqual(compare.check_strings(['ab', 'c'], ['ab', 'c']),
                         (0, []))

    def test_strings_fail(self):
        compare = FilesComparison()
        self.assertEqual(compare.check_strings([], ['x']),
                         (1, ['Strings have different numbers of lines',
                              'No files']))
        self.assertEqual(compare.check_strings(['y'], ['x']),
                         (1, ['1 line is different, starting at line 1',
                              'No files']))

    def test_print(self):
        msgs = []
        compare = FilesComparison(print_fn=lambda x: msgs.append(x))
        compare.check_strings(['a'], ['b'])
        self.assertEqual(msgs, ['1 line is different, starting at line 1',
                                'No files'])

    def test_strip(self):
        compare = FilesComparison()
        self.assertEqual(compare.check_strings(['   abc'], ['abc']),
                         (1, ['1 line is different, starting at line 1',
                              'No files']))
        self.assertEqual(compare.check_strings(['   abc'], ['abc'],
                                               lstrip=True), (0, []))
        self.assertEqual(compare.check_strings(['abc   '], ['abc'],
                                               rstrip=True), (0, []))
        self.assertEqual(compare.check_strings(['   abc   '], ['abc'],
                                               lstrip=True, rstrip=True),
                         (0, []))

    def test_ignore_substrings(self):
        compare = FilesComparison()
        self.assertEqual(compare.check_strings(['abc','red', 'banana'],
                                               ['abc','blue', 'grapefruit']),
                         (1, ['2 lines are different, starting at line 2',
                              'No files']))
        self.assertEqual(compare.check_strings(['abc','red', 'banana'],
                                               ['abc','blue', 'grapefruit'],
                                               ignore_substrings=['re']),
                         (1, ['1 line is different, starting at line 3',
                              'No files',
                              'Note exclusions:', '    re']))
        self.assertEqual(compare.check_strings(['abc','red', 'banana'],
                                               ['abc','blue', 'grapefruit'],
                                               ignore_substrings=['ue','gra']),
                         (0, []))

    def test_ignore_patterns(self):
        compare = FilesComparison()
        self.assertEqual(compare.check_strings(['abc','red', 'banana'],
                                               ['abc','blue', 'grapefruit']),
                         (1, ['2 lines are different, starting at line 2',
                              'No files']))
        self.assertEqual(compare.check_strings(['abc','red', 'banana'],
                                               ['abc','blue', 'grapefruit'],
                                               ignore_patterns=['gr.*t']),
                         (1, ['2 lines are different, starting at line 2',
                              'No files',
                              'Note exclusions:', '    gr.*t']))
        self.assertEqual(compare.check_strings(['abc','spangle', 'breadfruit'],
                                               ['abc','spanner', 'grapefruit'],
                                               ignore_patterns=['sp.....',
                                                                '.*fruit']),
                         (0, []))

    def test_preprocess(self):
        compare = FilesComparison()
        def strip_first_five(strings):
            return [s[5:] for s in strings]
        def strip_first_seven(strings):
            return [s[7:] for s in strings]
        self.assertEqual(compare.check_strings(['abc','spangle', 'breadfruit'],
                                               ['abc','spanner', 'grapefruit'],
                                               preprocess=strip_first_five),
                         (1, ['1 line is different, starting at line 2',
                              'No files']))
        self.assertEqual(compare.check_strings(['abc','spangle', 'breadfruit'],
                                               ['abc','spanner', 'grapefruit'],
                                               preprocess=strip_first_seven),
                         (0, []))

    def test_permutations(self):
        compare = FilesComparison()
        self.assertEqual(compare.check_strings(['abc','spangle', 'spanner'],
                                               ['spangle','spanner', 'abc'],
                                               max_permutation_cases=1),
                         (1, ['3 lines are different, starting at line 1',
                              'No files']))
        self.assertEqual(compare.check_strings(['abc','spangle', 'spanner'],
                                               ['abc','spanner', 'spangle'],
                                               max_permutation_cases=2),
                         (0, []))
        self.assertEqual(compare.check_strings(['abc','spangle', 'spanner'],
                                               ['spangle','spanner', 'abc'],
                                               max_permutation_cases=3),
                         (0, []))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""
examples.py: copy example files to a specified directory

Source repository: http://github.com/tdda/tdda

License: MIT

Copyright (c) Stochastic Solutions Limited 2016-2017
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os
import shutil
import sys


def copy_examples(name, destination='.', verbose=True):
    """
    Copy example files to a specified directory
    """
    path = os.path.join(os.path.dirname(__file__), name)
    srcdir = os.path.join(path, 'examples')
    if not os.path.isdir(destination):
        print('copyexamples: output directory %s does not exist' % destination,
              file=sys.stderr)
        sys.exit(1)
    outdir = os.path.join(destination, '%s-examples' % name)
    shutil.rmtree(outdir, ignore_errors=True)
    os.mkdir(outdir)
    copy(srcdir, outdir)
    if verbose:
        print('Copied example files for tdda.%s to %s' % (name, outdir))


def copy(srcdir, destination):
    """
    Recursive copy
    """
    for name in os.listdir(srcdir):
        fullname = os.path.join(srcdir, name)
        if os.path.isdir(fullname):
            if name in ('.cache', '__pycache__'):
                continue
            outdir = os.path.join(destination, name)
            if not os.path.exists(outdir):
                os.mkdir(outdir)
            copy(fullname, outdir)
        elif name.endswith('.pyc'):
            continue
        else:
            binary = 'b' if fullname.endswith('.feather') else ''
            with open(fullname, 'r%s' % binary) as fin:
                with open(os.path.join(destination, name),
                                       'w%s' % binary) as fout:
                    fout.write(fin.read())


def copy_main(name, verbose=True):
    if len(sys.argv) > 2:
        print('USAGE: examples [destination-directory]', file=sys.stderr)
        sys.exit(1)
    destination = sys.argv[1] if len(sys.argv) == 2 else '.'
    copy_examples(name, destination, verbose=verbose)




# -*- coding: utf-8 -*-

"""
Classes for representing individual constraints.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import datetime
import json
import os
import re
import sys

from collections import OrderedDict

PRECISIONS = ('open', 'closed', 'fuzzy')
STANDARD_FIELD_CONSTRAINTS = ('type', 'min', 'min_length', 'max', 'max_length',
                              'sign', 'max_nulls', 'no_duplicates',
                              'allowed_values', 'rex')
STANDARD_FIELD_GROUP_CONSTRAINTS = ('lt', 'lte', 'eq', 'gt', 'gte')
SIGNS = ('positive', 'non-negative', 'zero', 'non-positive', 'negative',
         'zero', 'null')
TYPES = ('bool', 'int', 'real', 'date', 'string')
DATE_VALUED_CONSTRAINTS = ('min', 'max')
UTF8 = 'UTF-8'


RD = re.compile(r'^(\d{4})[-/](\d{1,2})[-/](\d{1,2})$')
RDT = re.compile(r'^(\d{4})[-/](\d{1,2})[-/](\d{1,2})[ T]'
                 r'(\d{1,2}):(\d{2}):(\d{2})$')
RDTM = re.compile(r'^(\d{4})[-/](\d{1,2})[-/](\d{1,2})[ T]'
                  r'(\d{1,2}):(\d{2}):(\d{2})'
                  r'\.(\d+)$')

UNICODE_TYPE = str if sys.version_info[0] >= 3 else unicode

EPSILON_DEFAULT = 0.0   # no tolerance for min/max constraints for
                        # real (i.e. floating point) fields.


class Marks:
    tick = '✓'     # This is a tick mark; whether or not it displays in editors
    cross = '✗'    # This is a cross; again, it may not display
    nothing = '-'  # This is an en-dash; again, it may not display in editors


class SafeMarks:
    tick = 'OK'
    cross = 'X'
    nothing = '-'


class TDDAObject(OrderedDict):
    """
    Ordered Dictionary
    """
    def __init__(self, *args, **kwargs):
        OrderedDict.__init__(self, *args, **kwargs)

    def set_key_order(self, keys):
        keys = keys + [k for k in self.keys() if k not in keys]
        for k in keys:
            try:
                v = self[k]
                del self[k]
                self[k] = v
            except KeyError:
                pass


class DatasetConstraints(object):
    """
    Container for constraints pertaining to a dataset.
    Currently only supports per-field constraints.
    """
    def __init__(self, per_field_constraints=None, loadpath=None):
        if loadpath:
            self.fields = Fields()
            self.load(loadpath)
        else:
            self.fields = Fields(per_field_constraints)

    def __getitem__(self, k):
        if type(k) == int:
            k = self.fields.keys()[k]
        return self.fields[k]

    def __contains__(self, k):
        return k in self.fields

    def add_field(self, fc):
        self.fields[fc.name] = fc

    def __str__(self):
        return 'FIELDS:\n\n%s' % str(self.fields)

    def load(self, path):
        """
        Builds a DatasetConstraints object from a json file
        """
        with open(path) as f:
            text = f.read()
        self.initialize_from_dict(native_definite(json.loads(text)))

    def initialize_from_dict(self, in_constraints):
        """
        Initializes this object from a dictionary in_constraints.
        Currently, the only key used from in_constraints is fields.

        The value of in_constraints['fields'] is expected to be
        a dictionary, keyed on field name, whose values are the
        constraints for that field.

        They constraints are keyed on the kind of constraint, and should
        contain either a single value (a scalar or a list), or a dictionary
        of keyword arguments for the constraint initializer.
        """
        fields = in_constraints['fields'] or []
        for fieldname, c in fields.items():
            fc = []
            is_date = 'type' in c and c['type'] == 'date'
            for kind, value in c.items():
                constraint_constructor = FIELD_CONSTRAINTS_MAP.get(kind)
                if constraint_constructor:
                    if isinstance(value, dict):
                        constraint = constraint_constructor(**value)
                    else:
                        constraint = constraint_constructor(value)
                    if is_date and kind in DATE_VALUED_CONSTRAINTS:
                        constraint.value = get_date(constraint.value)
                    fc.append(constraint)
                else:
                    warn('Constraint kind %s for field %s unknown: ignored.'
                         % (kind, fieldname))
            if fc:
                self.add_field(FieldConstraints(fieldname, fc))

    def to_dict(self):
        """
        Converts the constraints in this object to a dictionary.
        """
        constraints = OrderedDict((
            (f, v.to_dict_value()) for f, v in self.fields.items()
        ))
        return OrderedDict((('fields', constraints),))

    def to_json(self):
        """
        Converts the constraints in this object to JSON.
        The resulting JSON is returned.
        """
        return json.dumps(self.to_dict(), indent=4) + '\n'

    def sort_fields(self, fields=None):
        """
        Sorts the field constraints within the object by field order,
        by default by alphabetical order.

        If a list of field names is provided, then the fields will appear
        in that given order (with any additional fields appended at the end).
        """
        if fields is None:
            fields = sorted(self.fields.keys())
        self.fields.set_key_order(fields)


class Fields(TDDAObject):
    def __init__(self, constraints=None):
        TDDAObject.__init__(self)
        for c in constraints or []:
            self[c.name] = c

    def to_dict_value(self):
        return OrderedDict((name, c.to_dict_value())
                             for (name, c) in self.items())

    def __str__(self):
        return str('\n\n'.join(str(v) for v in self.values()))


class FieldConstraints(object):
    """
    Container for constraints on a field.
    """
    def __init__(self, name=None, constraints=None):
        """
        The name of the field can be supplied, or left as null (None).
        Leaving it null can be appropriate if the same constraint is
        to be used for multiple fields.

        Constraints can be supplied as a list; if so, these will be copied
        into the dictionary using the constraint kind as a key.
        """
        self.name = name
        self.constraints = {}
        for c in constraints or []:
            self.constraints[c.kind] = c


    def to_dict_value(self):
        """
        Returns a pair consisting of the name supplied, or the stored name,
        and an ordered dictionary keyed on constraint kind with the value
        specifying the constraint. For simple constraints, the value is a
        base type; for more complex constraints with several components,
        the value will itself be an (ordered) dictionary.

        The ordering is all to make the JSON file get written in a sensible
        order, rather than being a jumbled mess.
        """
        d = OrderedDict()
        keys = to_preferred_order(self.constraints.keys(),
                                  STANDARD_FIELD_CONSTRAINTS)
        for k in keys:
            d[k] = self.constraints[k].to_dict_value()
        return d

    def __getitem__(self, k):
        if type(k) == int:
            keys = list(self.to_dict_value().keys())
            k = keys[k]
        return self.constraints[k]

    def __str__(self):
        keys = [k for k in STANDARD_FIELD_CONSTRAINTS
                if k in self.constraints]
        keys += list(sorted(set(self.constraints.keys()) - set(keys)))
        return str('Field %s:\n  %s' % (self.name,
                                        '\n  '.join('%13s: %s'
                                                    % (k, self.constraints[k])
                                                       for k in keys)))


class MultiFieldConstraints(FieldConstraints):
    """
    Container for constraints on a pairs (or higher numbers) of fields
    """
    def __init__(self, names=None, constraints=None):
        """
        The names of the fields can be supplied, or left as null (None).
        Leaving them null can be appropriate if the same constraint is
        to be used for multiple field groups, though will not serialize
        terrible well.

        Constraints can be supplied as a list; if so, these will be copied
        into the dictionary using the constraint kind as a key.
        """
        self.names = tuple(names)
        self.constraints = {}
        for c in constraints or []:
            self.constraints[c.kind] = c


    def to_dict_value(self):
        """
        Returns a pair consisting of
            - a comma-separated list of the field names
            - an ordered dictionary keyed on constraint kind with the value
              specifying the constraint.

        For simple constraints, the value is a
        base type; for more complex constraints with several components,
        the value will itself be an (ordered) dictionary.

        The ordering is all to make the JSON file get written in a sensible
        order, rather than being a jumbled mess.
        """
        d = OrderedDict()
        for k in STANDARD_FIELD_GROUP_CONSTRAINTS:
            if k in self.constraints:
                d[k] = self.constraints[k].to_dict_value()
        remainder = sorted(set(self.constraints.keys())
                           - set(STANDARD_FIELD_GROUP_CONSTRAINTS))
        for k in remainder:
            d[k] = self.constraints[k].to_dict_value()
        return d

    def __getitem__(self, k):
        if type(k) == int:
            keys = list(self.to_dict_value().keys())
            k = keys[k]
        return self.constraints[k]

    def name_key(self):
        return ','.join(self.names)

    def __str__(self):
        keys = [k for k in STANDARD_FIELD_GROUP_CONSTRAINTS
                if k in self.constraints]
        keys += list(sorted(set(self.constraints.keys()) - set(keys)))
        return str('Field %s:\n  %s' % (self.name_key(),
                                        '\n  '.join('%13s: %s'
                                                    % (k, self.constraints[k])
                                                       for k in keys)))

class Constraint(object):
    """
    Base container for a single constraint.
    All specific constraint types (should) subclass this.
    """
    def __init__(self, kind, value, **kwargs):
        """
        All constraints have a kind (a string) and a value, which should
        be a base type. The convention is that the value is a simple type,
        and that if it is null (None) the constraint is always satisifed.
        Some constraints do not really need a value (e.g., no_nulls),
        and in such cases the convention is the set the value to True
        to indicate that the constraint is in force.

        Some constraints accept or require extra parameters.
        These are supplied through keyword arguments.
        """
        self.kind = kind
        self.value = value
        for (k, v) in kwargs.items():
            self.__dict__[k] = v

        assert constraint_class(kind) == self.__class__.__name__

    def __repr__(self):
        """
        """
        kws = ', '.join('%s=%s' % (k, repr(v))
                        for (k, v) in sorted(self.__dict__.items())
                        if k not in ('kind', 'value'))
        return '%s(value=%s%s)' % (constraint_class(self.kind),
                                   repr(self.value),
                                   (', ' + kws) if kws else '')

    def to_dict_value(self):
        return (self.value if type(self.value) != datetime.datetime
                else str(self.value))


#
# SINGLE FIELD CONSTRAINTS
#

class MinConstraint(Constraint):
    """
    Constraint specifying the minimum allowed value in a field.
    """
    def __init__(self, value, precision=None):
        assert precision is None or precision in PRECISIONS
        Constraint.__init__(self, 'min', value, precision=precision)

    def to_dict_value(self):
        if self.precision is None:
            return Constraint.to_dict_value(self)
        else:
            return OrderedDict((('value', self.value),
                                ('precision', self.precision)))


class MaxConstraint(Constraint):
    """
    Constraint specifying the maximum allowed value in a field.
    """
    def __init__(self, value, precision=None):
        assert precision is None or precision in PRECISIONS
        Constraint.__init__(self, 'max', value, precision=precision)

    def to_dict_value(self):
        if self.precision is None:
            return Constraint.to_dict_value(self)
        else:
            return OrderedDict((('value', self.value),
                                ('precision', self.precision)))


class SignConstraint(Constraint):
    """
    Constraint specifying allowed sign of values in a field.
    Used only for numeric fields (``real``, ``int``, ``bool``), and normally
    used in addition to Min and Max constraints.

    Possible values are ``positive``, ``non-negative``, ``zero``,
    ``non-positive``, ``negative`` and ``null``.
    """
    def __init__(self, value):
        assert value is None or value in SIGNS
        Constraint.__init__(self, 'sign', value)


class TypeConstraint(Constraint):
    """
    Constraint specifying the allowed (TDDA) type of a field.
    This can be a single value, chosen from:

        - ``bool``
        - ``int``
        - ``real``
        - ``string``
        - ``date``

    or a list of such values, most commonly ``['int', 'real']``,
    sometimes used because of Pandas silent and automatic promotion
    of integer fields to floats if nulls are present.)
    """
    def __init__(self, value):
        if type(value) in (list, tuple):
            assert all(t in TYPES for t in value)
        else:
            assert value is None or value in TYPES
        Constraint.__init__(self, 'type', value)


class MaxNullsConstraint(Constraint):
    """
    Constraint on the maximum number of nulls allowed in a field.
    Usually 0 or 1.
    (The constraint generator only generates 0 and 1, but the verifier
    will verify and number.)
    """
    def __init__(self, value):
        Constraint.__init__(self, 'max_nulls', value)


class NoDuplicatesConstraint(Constraint):
    """
    Constraint specifying that non dupicate non-null values are allowed
    in a field.

    Currently only generated for string fields, though could be used
    more broadly.
    """
    def __init__(self, value=True):
        assert value is None or value == True
        Constraint.__init__(self, 'no_duplicates', value)


class AllowedValuesConstraint(Constraint):
    """
    Constraint restricting the allowed values in a field to an explicity list.

    Currently only used for string fields.

    When generating constraints, this code will only generate such a
    constraint if there are no more than ``MAX_CATEGORIES`` (= 20 at the
    time of writing, but check above in case this comment rusts)
    different values in the field.
    """
    def __init__(self, value):
        Constraint.__init__(self, 'allowed_values', value)


class MinLengthConstraint(Constraint):
    """
    Constraint restricting the minimum length of strings in a string field.

    Generated instead of a ``MinConstraint`` by this generation code,
    but can be used in conjunction with a ``MinConstraint``.
    """
    def __init__(self, value):
        Constraint.__init__(self, 'min_length', value)


class MaxLengthConstraint(Constraint):
    """
    Constraint restricting the maximum length of strings in a string field.

    Generated instead of a ``MaxConstraint`` by this generation code,
    but can be used in conjunction with a ``MinConstraint``.
    """
    def __init__(self, value):
        Constraint.__init__(self, 'max_length', value)


class RexConstraint(Constraint):
    """
    Constraint restricting a string field to match (at least) one of
    the regular expressions in a list given.
    """
    def __init__(self, value):
        Constraint.__init__(self, 'rex', [native_definite(v) for v in value])


#
# MULTI-FIELD CONSTRAINTS
#


class LtConstraint(Constraint):
    """
    Constraint specifying that the first field of a pair should be
    (strictly) less than the second, where both are non-null.
    """
    def __init__(self, value):
        Constraint.__init__(self, 'lt', value)


class LteConstraint(Constraint):
    """
    Constraint specifying that the first field of a pair should be
    no greater than the second, where both are non-null.
    """
    def __init__(self, value):
        Constraint.__init__(self, 'lte', value)

class EqConstraint(Constraint):
    """
    Constraint specifying that two fields should have identical values
    where they are both non-null.
    """
    def __init__(self, value):
        Constraint.__init__(self, 'eq', value)


class GtConstraint(Constraint):
    """
    Constraint specifying that the first field of a pair should be
    (strictly) greater than the second, where both are non-null.
    """
    def __init__(self, value):
        Constraint.__init__(self, 'gt', value)


class GteConstraint(Constraint):
    """
    Constraint specifying that the first field of a pair should be
    greater than or equal to the second, where both are non-null.
    """
    def __init__(self, value):
        Constraint.__init__(self, 'gte', value)


class Verification(object):
    """
    Container for the result of a constraint verification for a dataset
    in the context of a given set of constraints.
    """
    def __init__(self, constraints, report='all', one_per_line=False,
                 ascii=False, detect=False, detect_outpath=None,
                 detect_write_all=False, detect_per_constraint=False,
                 detect_output_fields=None, detect_rownumber=False,
                 detect_in_place=False):
        self.fields = TDDAObject()
        self.failures = 0
        self.passes = 0
        self.report = report
        self.one_per_line = one_per_line
        self.ascii = ascii
        self.detect = detect
        self.detect_outpath = detect_outpath
        self.detect_write_all = detect_write_all
        self.detect_per_constraint = detect_per_constraint
        self.detect_output_fields = detect_output_fields
        self.detect_rownumber = detect_rownumber
        self.detect_in_place = detect_in_place
        if report not in ('all', 'fields', 'constraints'):
            raise Exception('Value for report must be one of "all", "fields"'
                            ' or "constraints", not "%s".' % report)
        if not detect_outpath and not detect and not detect_in_place:
            if any((detect_write_all, detect_per_constraint,
                    detect_output_fields, detect_rownumber)):
                raise Exception('You have specified detection parameters '
                                'without specifying\na detection output path.')

    def __str__(self):
        """
        Returns string representation of the :py:class:`Verification` object.

        The format of the string is controlled by the value of the
        object's :py:attr:`report` property. If this is set to 'fields',
        then it reports only those fields that have failures.
        """
        if self.report == 'fields':  # Report only fields with failures
            field_items = list((field, ver)
                               for (field, ver) in self.fields.items()
                               if ver.failures > 0)
        else:
            field_items = self.fields.items()
        fields = '\n\n'.join('%s: %s  %s  %s'
                           % (field,
                              plural(ver.failures, 'failure'),
                              plural(ver.passes, 'pass', 'es'),
                              '  '.join('%s %s' % (c, tcn(s, self.ascii))
                                       for (c, s) in ver.items()))
                           for field, ver in field_items)
        fields_part = 'FIELDS:\n\n%s\n\n' % fields if fields else ''

        #TODO: could easily report number of passing/failing fields too.
        #TODO: could (with more difficulty) report number of p/f records too.

        return ('%sSUMMARY:\n\n'
                'Constraints passing: %d\n'
                'Constraints failing: %d'
                % (fields_part, self.passes, self.failures))


def constraint_class(kind):
    """
    The convention is that name of the class implementing the constraint
    has a simple relationship to the constraint kind, namely that
    a constraint whose kind is 'this_kind' is implemented by a class
    called ThisKindConstraint.

    So:

        ``min``      --> ``MinConstraint``
        ``min_length``  --> ``MinLengthConstraint``
        ``no_nulls`` --> ``NoNullsConstraint``

    etc.

    This function maps the constraint kind to the class name using this rule.
    """
    return '%sConstraint' % ''.join(part.title() for part in kind.split('_'))


def strip_lines(s, side='r'):
    """
    Splits the given string into lines (at newlines), strips each line
    and rejoins. Is careful about last newline and default to stripping
    on the right only (side='r'). Use 'l' for left strip or anything
    else to strip both sides.
    """
    strip = (str.rstrip if side == 'r' else str.lstrip if side == 'l'
             else str.strip)
    end = '\n' if s.endswith('\n') else ''
    return '\n'.join([strip(line) for line in s.splitlines()]) + end


def verify(constraints, fieldnames, verifiers, VerificationClass=None,
           detected_records_writer=None, **kwargs):
    """
    Perform a verification of a set of constraints.
    This is primarily an internal function, intended to be used by
    specific verifiers for various types of data.
    (Specifically, at the moment, the Pandas verifier verify_df
    uses this function.)

    Inputs:

        constraints         is a DatasetConstraints object.

        verifiers           is a mapping from constraint kind to a verifier
                            callable.

                            NOTE: normally the verifier callable is a method
                            on a class that "knows" the dataset to be verified.

        VerificationClass   If provided, this should be a subclass of
                            Verificatation. This option is provided so
                            that callers can get back Verification object
                            with extra convenience methods. For example,
                            The Pandas code passes in a PandasVerification
                            class, which as a to_frame() method for turning
                            the result of the verification into a Pandas
                            DataFrame. If not provided, Verification
                            is used.

        kwargs              Any keyword arguments provided are passed to
                            the VerificationClass chosen.

                            report, one_per_line and ascii can all be set
                            this way.

    Returns a Verification object.
    """
    VerificationClass = VerificationClass or Verification
    results = VerificationClass(constraints, **kwargs)
    detect_outpath = kwargs.get('detect_outpath')
    detect = (detect_outpath is not None
              or kwargs.get('detect') is not None
              or kwargs.get('detect_in_place') is not None)
    allfields = sorted(constraints.fields.keys(),
                       key=lambda f: fieldnames.index(f) if f in fieldnames
                                                         else -1)

    if detect_outpath:
        # empty (and then remove) the detection output file first,
        # so that we can get an early error if the file isn't writeable,
        # and so that we don't leave a bogus wrong file in place if
        # we turn out not to detect anything.
        with open(detect_outpath, 'w') as f:
            pass
        os.remove(detect_outpath)

    for name in allfields:
        field_results = TDDAObject()
        failures = passes = 0
        for c in constraints.fields[name]:
            verify = verifiers.get(c.kind)
            if verify:
                satisfied = verify(name, c, detect)
                if satisfied:
                    passes += 1
                else:
                    failures += 1
            else:
                satisfied = None
            field_results[c.kind] = satisfied
        field_results.failures = failures
        results.failures += failures
        field_results.passes = passes
        results.passes += passes
        results.fields[name] = field_results

    if detect and detected_records_writer and results.failures > 0:
        results.detection = detected_records_writer(**kwargs)
    return results


def tcn(sat, ascii=False):
    """
    Convert True/False/None value to the appropriate tick, cross
    or nothing mark for printing.
    """
    marks = SafeMarks if ascii else Marks
    return marks.nothing if sat is None else marks.tick if sat else marks.cross


def warn(s):
    print(s, file=sys.stderr)



#
# Mapping from constraint kind (e.g. 'min_length') to constraint class
# e.g. MinLengthConstraint.
#
# Note: Each mapped value is the class iself, not its name.
#

FIELD_CONSTRAINTS_MAP = {
    kind: eval(constraint_class(kind)) for kind in STANDARD_FIELD_CONSTRAINTS
}


def plural(n, s, pl=None):
    """
    Returns a string like '23 fields' or '1 field' where the
    number is n, the stem is s and the plural is either stem + 's'
    or stem + pl (if provided).
    """
    if pl is None:
        pl = 's'
    if n == 1:
        return '%s %s' % (n, s)
    else:
        return '%s %s%s' % (n, s, pl)


def native_definite(o):
    return (UnicodeDefinite(o) if sys.version_info[0] >= 3
                               else UTF8DefiniteObject(o))


def UTF8Definite(s):
    """
    Converts a string to UTF-8 if it is unicode.
    Otherwise just returns the string.
    """
    return s if type(s) == bytes else s.encode(UTF8)


def UnicodeDefinite(s):
    """
    Converts a string to unicode if it is unicode.
    Otherwise just returns the string.
    """
    return s.decode(UTF8) if type(s) == bytes else s


def UTF8DefiniteObject(s):
    """
    Converts all unicode within scalar or object, recursively, to unicode.
    Handles lists, tuples and dictionaries, as well as scalars.
    """
    if type(s) == UNICODE_TYPE:
        return s.encode(UTF8)
    elif type(s) == list:
        return [UTF8DefiniteObject(v) for v in s]
    elif type(s) == tuple:
        return tuple([UTF8DefiniteObject(v) for v in s])
    elif type(s) == dict:
        return {UTF8DefiniteObject(k): UTF8DefiniteObject(v)
                for (k, v) in s.items()}
    return s


def get_date(d):
    for rex, L in ((RD, 3), (RDT, 6), (RDTM, 7)):
        m = re.match(rex, d)
        if m:
            try:
                return datetime.datetime(*(int(m.group(i))
                                           for i in range(1, L + 1)))
            except ValueError:
                print ('Failed to read "%s" as date' % d, file=sys.stderr)
                return d
    return d


def to_preferred_order(keys, preferred_order):
    return ([k for k in preferred_order if k in list(keys)]
               + list(sorted(set(keys) - set(preferred_order))))



def fuzzy_greater_than(a, b, epsilon):
    """
    Returns a >~ b (a is greater than or approximately equal to b)

    At the moment, this simply reduces b by 1% if it is positive,
    and makes it 1% more negative if it is negative.
    """
    return (a >= b) or (a >= fuzz_down(b, epsilon))


def fuzzy_less_than(a, b, epsilon):
    """
    Returns a <~ b (a is less than or approximately equal to b)

    At the moment, this increases b by 1% if it is positive,
    and makes it 1% less negative if it is negative.
    """
    return (a <= b) or (a <= fuzz_up(b, epsilon))


def fuzz_down(v, epsilon):
    """
    Adjust v downwards, by a proportion controlled by self.epsilon.
    This is typically used for fuzzy minimum constraints.

    By default, positive values of v are reduced by 1% so that slightly
    smaller values can pass the fuzzy minimum constraint.

    Similarly, negative values are made 1% more negative, so that
    slightly more negative values can still pass a fuzzy minimum
    constraint.
    """
    return v * ((1 - epsilon) if v >= 0 else (1 + epsilon))


def fuzz_up(v, epsilon):
    """
    Adjust v upwards, by a proportion controlled by self.epsilon.
    This is typically used for fuzzy maximum constraints.

    By default, positive values of v are increased by 1% so that
    slightly larger values can pass the fuzzy maximum constraint.

    Similarly, negative values are made 1% less negative, so that
    slightly less negative values can still pass a fuzzy maximum
    constraint.
    """
    return v * ((1 + epsilon) if v >= 0 else (1 - epsilon))


def sort_constraint_dict(d):
    """
    Helper function for tests, to sort a constraints dictionary (read
    from a .tdda file) into alphabetical order by field name, and with
    all of the individual constraints in the same order in which they
    are generated.
    """
    constraintkey = ['type', 'min', 'max', 'min_length', 'max_length',
                     'sign', 'max_nulls', 'no_duplicates', 'allowed_values',
                     'rex']
    fields = OrderedDict((
        (f, OrderedDict(((k, kv)
                         for k, kv in sorted(v.items(),
                                             key=lambda x:
                                                  constraintkey.index(x[0])))))
        for f, v in sorted(d['fields'].items())
    ))
    return OrderedDict((('fields', fields),))




# elements_constraints_discovery.py

from __future__ import print_function
import pandas as pd

from tdda.constraints.pd.constraints import discover_df

df = pd.read_csv('testdata/elements92.csv')
constraints = discover_df(df)
with open('elements92.tdda', 'w') as f:
    f.write(constraints.to_json())
print('Written elements92.tdda')





# -*- coding: utf-8 -*-
"""
generators.py: Trivial result generation functions for illustrating
               tdda.referencetest

Source repository: http://github.com/tdda/tdda

License: MIT

Copyright (c) Stochastic Solutions Limited 2016
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import math

def generate_string():
    """
    Returns an HTML string that should display identically to the result
    stored in tdda/examples/reference/string_result, but which has
    four (non-semantic) differences:
        - The Copyright line is different
        - The version number line is different
        - It has a newline at the top, unlike the reference file.
        - It is missing a newline at the end of the string, which the
          reference file includes.
    """
    return """<!DOCTYPE html>
<html>
  <head>
    <!--
    Copyright (c) Stochastic Solutions, 2016
    Version 1.0.0
    -->
    <meta charset="UTF-8"/>
    <style type="text/css">
      body {font-family: Palatino, "Palatino Linotype", Times;
            text-align: center}
      h1 {font-size: large; text-align: center;}
      div {padding: 12px 0 12px 0;}
    </style>
    <title>
      Python-Generated HTML Example for tdda.referencetest
    </title>
  </head>
<body>
  <h1>Python-Generated HTML Example for tdda.referencetest</h1>
  <div>
    This page is generated by Python (as a string).
  </div>
  %s
  <div>
    It's not terribly exciting.
    But it will serve to illustrate tdda.referencetest.
  </div>
</body>
</html>
""" % generate_spiral()


def generate_file(path):
    html = '''<!DOCTYPE html>
<html>
  <head>
    <!--
    Copyright (c) Stochastic Solutions, 2016
    Version 1.0.0
    -->
    <meta charset="UTF-8"/>
    <style type="text/css">
      body {font-family: Palatino, "Palatino Linotype", Times;
            text-align: center;}
      h1 {font-size: large; text-align: center;}
      div {padding-top: 12px; padding-bottom: 12px;}
    </style>
    <title>
      Python-Generated HTML Example for tdda.referencetest
    </title>
  </head>
<body>
  <h1>Python-Generated HTML Example for tdda.referencetest</h1>
  <div>
    This page is generated by Python (as a file).
  </div>
  <svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="200" height="200">
    <style type="text/css">text { font-family: Helvetica, Verdana, Sans;}</style>

    <!-- TDDA Test Output Version 0.0.1
         Copyright (c) Stochastic Solutions Limited 2016 -->

    <rect x="0" y="0" width="96" height="96" style="fill:#C08080"/>
    <rect x="105" y="0" width="96" height="96" style="fill:#F0C080"/>
    <rect x="0" y="104" width="96" height="96" style="fill:#F0C080"/>
    <rect x="104" y="104" width="96" height="96" style="fill:#C08080"/>
    <text x="50" y="78" font-size="80" text-anchor="middle">T</text>
    <text x="150" y="78" font-size="80" text-anchor="middle">D</text>
    <text x="50" y="178" font-size="80" text-anchor="middle">D</text>
    <text x="150" y="178" font-size="80" text-anchor="middle">A</text>
  </svg>

  <div>
    It will serve to illustrate tdda.referencetest.
  </div>
</body>
</html>
'''
    with open(path, 'w') as f:
        f.write(html)


def generate_spiral():
    return '''<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="100" height="100">
    <g transform="translate(50,50)">
        <path d="M 0 0 %s z" stroke="red" fill="none"/>
    </g>
</svg>''' % ' '.join('%d %d L' % (t / 20 * math.cos(t * 2 * math.pi / 100),
                                  t / 20 * math.sin(t * 2 * math.pi / 100))
                     for t in range(1000))






# -*- coding: utf-8 -*-

#
# Unit tests for file functions from tdda.referencetest.checkfiles
#

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import division

import os
import unittest

from tdda.referencetest.checkfiles import FilesComparison
from tdda.referencetest.basecomparison import diffcmd


def refloc(filename):
    return os.path.join(os.path.dirname(__file__), 'testdata', filename)

def check(compare, values, filename, diff=False, actual_path=None):
    (code, errs) = compare.check_string_against_file(values, refloc(filename),
                                                     actual_path=actual_path)
    if not diff:
        errs = [e for e in errs if not e.startswith('    ' + diffcmd())]
        errs = [e for e in errs if not e.startswith('Compare with:')]
    return (code, errs)


class TestFiles(unittest.TestCase):

    def test_strings_against_files_ok(self):
        compare = FilesComparison()
        r1 = compare.check_string_against_file([], refloc('empty.txt'))
        r2 = compare.check_string_against_file('', refloc('empty.txt'))
        r3 = compare.check_string_against_file([''], refloc('empty.txt'))
        r4 = compare.check_string_against_file(['a single line'],
                                               refloc('single.txt'))
        self.assertEqual(r1, (0, []))
        self.assertEqual(r2, (0, []))
        self.assertEqual(r3, (0, []))
        self.assertEqual(r4, (0, []))

    def test_strings_against_files_fail(self):
        compare = FilesComparison()
        r1 = check(compare, ['x'], 'empty.txt')
        r2 = check(compare, 'x', 'empty.txt')
        r3 = check(compare, ['', ''], 'empty.txt')
        r4 = check(compare, ['the wrong text\n'], 'single.txt')
        r5 = check(compare, ['the wrong text\n'], 'single.txt', diff=True,
                   actual_path='wrong.txt')
        self.assertEqual(r1, (1, ['Strings have different numbers of lines',
                                  'Expected file %s' % refloc('empty.txt')]))
        self.assertEqual(r2, (1, ['Strings have different numbers of lines',
                                  'Expected file %s' % refloc('empty.txt')]))
        self.assertEqual(r3, (1, ['Strings have different numbers of lines',
                                  'Expected file %s' % refloc('empty.txt')]))
        self.assertEqual(r4, (1,
                              ['1 line is different, starting at line 1',
                               'Expected file %s' % refloc('single.txt')]))
        diff = '%s %s %s' % (diffcmd(), 'wrong.txt', refloc('single.txt'))
        self.assertEqual(r5, (1, ['1 line is different, starting at line 1',
                                  'Compare with:\n    %s\n' % diff]))

    def test_files_ok(self):
        compare = FilesComparison()
        r1 = compare.check_file(refloc('empty.txt'), refloc('empty.txt'))
        r2 = compare.check_file(refloc('single.txt'), refloc('single.txt'))
        r3 = compare.check_file(refloc('colours.txt'),
                                refloc('colours.txt'))
        self.assertEqual(r1, (0, []))
        self.assertEqual(r2, (0, []))
        self.assertEqual(r3, (0, []))

    def test_files_fail(self):
        compare = FilesComparison()
        r1 = compare.check_file(refloc('empty.txt'), refloc('single.txt'))
        r2 = compare.check_file(refloc('single.txt'), refloc('empty.txt'))
        r3 = compare.check_file(refloc('single.txt'),
                                refloc('colours.txt'))
        diff1 = '%s %s %s' % (diffcmd(),
                              refloc('empty.txt'), refloc('single.txt'))
        diff2 = '%s %s %s' % (diffcmd(),
                              refloc('single.txt'), refloc('empty.txt'))
        diff3 = '%s %s %s' % (diffcmd(),
                              refloc('single.txt'), refloc('colours.txt'))
        self.assertEqual(r1, (1, ['Files have different numbers of lines',
                                  'Compare with:\n    %s\n' % diff1]))
        self.assertEqual(r2, (1, ['Files have different numbers of lines',
                                  'Compare with:\n    %s\n' % diff2]))
        self.assertEqual(r3, (1, ['Files have different numbers of lines',
                                  'Compare with:\n    %s\n' % diff3]))

    def test_multiple_files_ok(self):
        compare = FilesComparison()
        r = compare.check_files([refloc('empty.txt'),
                                 refloc('single.txt'),
                                 refloc('colours.txt')],
                                [refloc('empty.txt'),
                                 refloc('single.txt'),
                                 refloc('colours.txt')])
        self.assertEqual(r, (0, []))

    def test_multiple_files_fail(self):
        compare = FilesComparison()
        r = compare.check_files([refloc('empty.txt'),
                                 refloc('single.txt'),
                                 refloc('colours.txt')],
                                [refloc('single.txt'),
                                 refloc('colours.txt'),
                                 refloc('colours.txt')])
        diff1 = '%s %s %s' % (diffcmd(),
                              refloc('empty.txt'), refloc('single.txt'))
        diff2 = '%s %s %s' % (diffcmd(),
                              refloc('single.txt'), refloc('colours.txt'))
        self.assertEqual(r, (2, ['Files have different numbers of lines',
                                 'Compare with:\n    %s\n' % diff1,
                                 'Files have different numbers of lines',
                                 'Compare with:\n    %s\n' % diff2]))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

#
# Unit tests for functions from tdda.referencetest.checkpandas
#

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import division

import os
import unittest

try:
    import pandas as pd
except ImportError:
    pd = None

from tdda.referencetest.checkpandas import PandasComparison
from tdda.referencetest.basecomparison import diffcmd

def refloc(filename):
    return os.path.join(os.path.dirname(__file__), 'testdata', filename)


@unittest.skipIf(pd is None, 'no pandas')
class TestStrings(unittest.TestCase):

    def test_frames_ok(self):
        compare = PandasComparison()
        df1 = pd.DataFrame({'a': [1, 2, 3, 4, 5],
                            'b': [1.0001, 2.0001, 3.0001, 4.0001, 5.0001]})
        df2 = pd.DataFrame({'a': [1, 2, 3, 4, 5],
                            'b': [1.0002, 2.0002, 3.0002, 4.0002, 5.0002]})
        df3 = pd.DataFrame({'a': [1, 2, 3, 4, 5],
                            'b': [1, 2, 3, 4, 5]})
        self.assertEqual(compare.check_dataframe(df1, df1), (0, []))
        self.assertEqual(compare.check_dataframe(df1, df2, precision=3),
                         (0, []))
        self.assertEqual(compare.check_dataframe(df1, df3,
                                                 check_types=['a'],
                                                 check_data=['a']),
                         (0, []))

    def test_frames_fail(self):
        compare = PandasComparison()
        df1 = pd.DataFrame({'a': [1, 2, 3, 4, 5],
                            'b': [1.0001, 2.0001, 3.0001, 4.0001, 5.0001]})
        df2 = pd.DataFrame({'a': [1, 2, 3, 4, 5],
                            'b': [1.0002, 2.0002, 3.0002, 4.0002, 5.0002]})
        df3 = pd.DataFrame({'a': [1, 2, 3, 4, 5],
                            'b': [1, 2, 3, 4, 5]})
        fromrow12 = ('From row 1: [1.000100, 2.000100, 3.000100, '
                     '4.000100, 5.000100] != [1.000200, 2.000200, '
                     '3.000200, 4.000200, 5.000200]')
        fromrow13 = ('From row 1: [1.000100, 2.000100, 3.000100, '
                     '4.000100, 5.000100] != [1, 2, 3, 4, 5]')
        self.assertEqual(compare.check_dataframe(df1, df2, precision=3),
                         (0, []))
        self.assertEqual(compare.check_dataframe(df1, df2, precision=6),
                         (1, ['Contents check failed.',
                              'Column values differ: b',
                              fromrow12]))
        self.assertEqual(compare.check_dataframe(df1, df3, check_types=['a'],
                                                 precision=3),
                         (1, ['Contents check failed.',
                              'Column values differ: b',
                              'Different types']))
        self.assertEqual(compare.check_dataframe(df1, df3, check_types=['a']),
                         (1, ['Contents check failed.',
                              'Column values differ: b',
                              fromrow13]))
        self.assertEqual(compare.check_dataframe(df1, df3, check_data=['a']),
                         (1,
                          ['Column check failed.',
                           'Wrong column type b (float64, expected int64)']))

    def test_pandas_csv_ok(self):
        compare = PandasComparison()
        r = compare.check_csv_file(refloc('colours.txt'),
                                   refloc('colours.txt'))
        self.assertEqual(r, (0, []))

    def test_pandas_csv_fail(self):
        compare = PandasComparison()
        (code, errs) = compare.check_csv_file(refloc('single.txt'),
                                              refloc('colours.txt'))
        errs = [e for e in errs if not e.startswith('Compare with:')
                                   and not e.startswith('    ' + diffcmd())]
        self.assertEqual(code, 1)
        self.assertEqual(errs,
                         ['Column check failed.',
                          'Missing columns: [%s]'
                              % ', '.join(["'%s'" % s for s in
                                           ['Name', 'RGB', 'Hue',
                                            'Saturation', 'Value']]),
                          'Extra columns: [\'a single line\']',
                          'Length check failed.',
                          'Found 0 records, expected 147'])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Run all the unit-tests for the referencetest module.
#

from __future__ import absolute_import
from __future__ import unicode_literals

import unittest

from tdda.referencetest.tests.teststrings import *
from tdda.referencetest.tests.testfiles import *
from tdda.referencetest.tests.testpandas import *
from tdda.referencetest.tests.testregeneration import *

if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-

"""
checkfiles.py: comparison mechanism for text files

Source repository: http://github.com/tdda/tdda

License: MIT

Copyright (c) Stochastic Solutions Limited 2016-2018
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
# from __future__ import unicode_literals

import os
import re
import sys
import tempfile

from tdda.referencetest.basecomparison import BaseComparison, copycmd


class FilesComparison(BaseComparison):

    def check_strings(self, actual, expected,
                      actual_path=None, expected_path=None,
                      lstrip=False, rstrip=False,
                      ignore_substrings=None, ignore_patterns=None,
                      ignore_lines=None,
                      preprocess=None, max_permutation_cases=0, msgs=None):
        """
        Compare two lists of strings (actual and expected), one-by-one.

            *actual*
                                is a list of strings.
            *expected*
                                is the expected list of strings.
            *actual_path*
                                is the path of the file where the actual
                                string originated (used for reporting errors).
            *expected_path*
                                is the path from which expected was read; it is
                                used to suggest a diff command to use on files
                                for strings failing the check.
            *lstrip*
                                if set to true, both strings are left-stripped
                                before the comparison is carried out.
                                Note: the stripping on a per-line basis.
            *rstrip*
                                if set to true, both strings are right-stripped
                                before the comparison is carried out.
                                Note: the stripping on a per-line basis.
            *ignore_substrings*
                                is an optional list of substrings; lines
                                containing any of these substrings will be
                                ignored in the comparison.
            *ignore_patterns*
                                is an optional list of regular expressions;
                                lines will be considered to be the same if
                                they only differ in substrings that match one
                                of these regular expressions. The expressions
                                must not contain parenthesised groups, and
                                should only include explicit anchors if they
                                need refer to the whole line.
            *ignore_lines*
                                is an optional list of substrings; lines
                                containing any of these substrings will be
                                completely removed before carrying out the
                                comparison. This is the means by which you
                                would exclude 'optional' content.
            *preprocess*
                                is an optional function that takes a list of
                                strings and preprocesses it in some way; this
                                function will be applied to both the actual
                                and expected.
            *max_permutation_cases*
                                is an optional number specifying the maximum
                                number of permutations allowed; if the actual
                                and expected lists differ only in that their
                                lines are permutations of each other, and
                                the number of such permutations does not
                                exceed this limit, then the two are considered
                                to be identical.
            *msgs*
                                is an optional list, where information about
                                differences will be appended; if not specified,
                                a new list will be created and returned.

        Returns a tuple (failures, msgs), where failures is 1 if the lists
        differ and 0 if they are the same. The returned msgs is a list
        containing information about how they differed.
        """

        if msgs is None:
            msgs = []
        first_error = None
        failure_cases = []

        if preprocess:
            expected = preprocess(expected)
            actual = preprocess(actual)

        if actual and len(actual[-1]) == 0:
            actual = actual[:-1]
        if expected and len(expected[-1]) == 0:
            expected = expected[:-1]

        if ignore_lines:
            actual = [a for a in actual
                      if not any(i in a for i in ignore_lines)]
            expected = [a for a in expected
                        if not any(i in a for i in ignore_lines)]

        if len(actual) == len(expected):
            normalize = self.normalize_function(lstrip, rstrip)
            diffs = [i for i in range(len(actual))
                     if normalize(actual[i]) != normalize(expected[i])]
            ndiffs = len(diffs)
            if ndiffs > 0:
                ignore_substrings = ignore_substrings or []
                ignore_patterns = ignore_patterns or []
                anchored_patterns = [('' if p.startswith('^') else '^(.*)')
                                      + p
                                      + ('' if p.endswith('$') else '(.*)$')
                                     for p in ignore_patterns]
                cPatterns = [re.compile(p) for p in anchored_patterns]
                if any(cp.groups > 3 for cp in cPatterns):
                    raise Exception('Invalid patterns: %s' % ignore_patterns)
                for i in diffs:
                    for pattern in ignore_substrings:
                        if pattern in actual[i] or pattern in expected[i]:
                            break
                    else:
                        # not an ignorable substring line, so try patterns
                        for pattern in cPatterns:
                            mExpected = re.match(pattern, expected[i])
                            if mExpected:
                                mActual = re.match(pattern, actual[i])
                                if not mActual:
                                    continue
                                if pattern.groups < 3:
                                    # matched an anchored expression
                                    break
                                lhs = mActual.group(1) + mActual.group(3)
                                rhs = (mExpected.group(1)
                                       + mExpected.group(3))
                                if lhs == rhs:
                                    actual[i] = (mActual.group(1) + '...'
                                                 + mActual.group(3))
                                    expected[i] = (mExpected.group(1)
                                                   + '...'
                                                   + mExpected.group(3))
                                    break
                        else:
                            # difference can't be ignored
                            if first_error is None:
                                first_error = (
                                    '%d line%s different, starting at line'
                                    ' %d'
                                    % (ndiffs,
                                       's are' if ndiffs != 1
                                               else ' is', i+1))
                            if len(failure_cases) < max_permutation_cases:
                                failure_cases.append((i,
                                                      actual[i],
                                                      expected[i]))
                            else:
                                break
                    # ignored a line, so there is one fewer to report
                    ndiffs -= 1
        else:
            ndiffs = max_permutation_cases + 1
            first_error = ('%s have different numbers of lines'
                           % ('Files' if actual_path else 'Strings'))

        if ndiffs > 0 and ndiffs <= max_permutation_cases:
            ndiffs = self.check_for_permutation_failures(failure_cases)
        if ndiffs > 0:
            if first_error:
                self.info(msgs, first_error)
            self.add_failures(msgs, actual_path, expected_path,
                              ignore_substrings=ignore_substrings,
                              ignore_patterns=ignore_patterns,
                              preprocess=preprocess, actual=actual,
                              expected=expected)
        return (1 if ndiffs > 0 else 0, msgs)

    def check_string_against_file(self, actual, expected_path,
                                  actual_path=None,
                                  lstrip=False, rstrip=False,
                                  ignore_substrings=None,
                                  ignore_patterns=None,
                                  ignore_lines=None,
                                  preprocess=None, max_permutation_cases=0,
                                  msgs=None):
        """
        Check a string (or list of strings) against the contents of a
        reference file.

        This is a wrapper around check_strings(), where the 'expected'
        strings are read from a file rather than being passed in explicitly.

        Other parameters are the same as for py:meth:`check_strings()`.

        The actual_path parameter is the pathname of the file that the
        actual string originally came from (if it came from a file at all;
        if not, it should be None).
        """
        if msgs is None:
            msgs = []
        try:
            with open(expected_path) as f:
                content = f.read()
                expected_ends_with_newline = content.endswith('\n')
                expected = content.splitlines()
        except IOError:
            self.info(msgs, 'Reference file %s not found.' % expected_path)
            self.add_failures(msgs, None, expected_path, actual=actual)
            return (1, msgs)

        if type(actual) in (list, tuple):
            actuals = actual
            actual_ends_with_newline = expected_ends_with_newline
        else:
            actuals = actual.splitlines()
            actual_ends_with_newline = actual.endswith('\n')
        (code, msgs) = self.check_strings(actuals, expected,
                                          actual_path=actual_path,
                                          expected_path=expected_path,
                                          lstrip=lstrip, rstrip=rstrip,
                                          ignore_substrings=ignore_substrings,
                                          ignore_patterns=ignore_patterns,
                                          ignore_lines=ignore_lines,
                                          preprocess=preprocess,
                                          max_permutation_cases=
                                              max_permutation_cases,
                                          msgs=msgs)
        #if expected_ends_with_newline != actual_ends_with_newline:
        #    code = 1
        #    if actual_ends_with_newline:
        #        self.info(msgs, 'Actual string has unexpected newline at end')
        #    else:
        #        self.info(msgs, 'Actual string is missing newline at end')
        return (code, msgs)

    def check_file(self, actual_path, expected_path,
                   lstrip=False, rstrip=False,
                   ignore_substrings=None, ignore_patterns=None,
                   ignore_lines=None,
                   preprocess=None, max_permutation_cases=0, msgs=None):
        """
        Check a pair of files, line by line, with optional
        ignore patterns (substrings) and optionally left-
        and/or right-stripping the contents of both files.

        This is a wrapper around check_strings(), where both the 'actual'
        and 'expected' strings are read from files rather than being passed
        in explicitly.

        Other parameters are the same as for :py:meth:`check_strings()`.

        """
        if msgs is None:
            msgs = []
        try:
            with open(expected_path) as f:
                content = f.read()
                expected_ends_with_newline = content.endswith('\n')
                expected = content.splitlines()
        except IOError:
            self.info(msgs, 'Reference file %s not found.' % expected_path)
            self.info(msgs,
                      'Initialize from actual content with:\n    %s %s %s'
                      % (copycmd(), actual_path, expected_path))
            return (1, msgs)
        try:
            with open(actual_path) as f:
                content = f.read()
                actual_ends_with_newline = content.endswith('\n')
                actuals = content.splitlines()
        except IOError:
            self.info(msgs, 'Actual file %s not found.'
                            % os.path.normpath(actual_path))
            self.add_failures(msgs, actual_path, expected_path)
            return (1, msgs)
        (code, msgs) = self.check_strings(actuals, expected,
                                          actual_path=actual_path,
                                          expected_path=expected_path,
                                          lstrip=lstrip, rstrip=rstrip,
                                          ignore_substrings=ignore_substrings,
                                          ignore_patterns=ignore_patterns,
                                          ignore_lines=ignore_lines,
                                          preprocess=preprocess,
                                          max_permutation_cases=
                                              max_permutation_cases,
                                          msgs=msgs)
        #if expected_ends_with_newline != actual_ends_with_newline:
        #    code = 1
        #    if actual_ends_with_newline:
        #        self.info(msgs, 'Actual string has unexpected newline at end')
        #    else:
        #        self.info(msgs, 'Actual string is missing newline at end')
        return (code, msgs)

    def check_files(self, actual_paths, expected_paths,
                    lstrip=False, rstrip=False,
                    ignore_substrings=None, ignore_patterns=None,
                    ignore_lines=None,
                    preprocess=None, max_permutation_cases=0, msgs=None):
        """
        Compare a list of files against a list of reference files.

        It compares all the files pair-wise from the two lists, and then
        reports on any differences. This is different from calling
        check_file() separately for each pair, which will stop as soon
        as the first difference is found.

        Other parameters are the same as for :py:meth:`check_strings()`.

        """
        failures = 0
        if msgs is None:
            msgs = []
        for (actual_path, expected_path) in zip(actual_paths, expected_paths):
            try:
                r = self.check_file(actual_path, expected_path,
                                    ignore_substrings=ignore_substrings,
                                    ignore_patterns=ignore_patterns,
                                    ignore_lines=ignore_lines,
                                    preprocess=preprocess,
                                    lstrip=lstrip, rstrip=rstrip,
                                    max_permutation_cases=max_permutation_cases,
                                    msgs=msgs)
                (n, msgs) = r
                failures += n
            except Exception as e:
                self.info(msgs, 'Error comparing %s and %s (%s %s)'
                                % (os.path.normpath(actual_path),
                                   expected_path,
                                   e.__class__.__name__, str(e)))
                failures += 1
        return (failures, msgs)

    def check_for_permutation_failures(self, failure_cases):
        """
        Check a collection of actual/reference comparison lines, to
        see if they are different only because one is a permutation of
        the other.
        """
        actuals = [f[1] for f in failure_cases]
        expecteds = [f[2] for f in failure_cases]
        if self.verbose and self.print_fn:
            self.print_fn('W(%s)' %len(actuals), file=sys.stderr)
        if sorted(actuals) == sorted(expecteds):
            return 0
        else:
            return len(actuals)

    def normalize_function(self, left, right):
        """
        Return the appropriate function for stripping a string,
        with left and right being booleans that specify whether
        to strip on the two sides.
        """
        if left and right:
            return lambda s: s.strip()
        elif left:
            return lambda s: s.lstrip()
        elif right:
            return lambda s: s.rstrip()
        else:
            return lambda s: s

    def add_failures(self, msgs, actual_path, expected_path,
                     ignore_substrings=None, ignore_patterns=None,
                     preprocess=None, actual=None, expected=None):
        """
        Build a list of messages describing the way in which two files are
        different.
        """
        if actual_path and expected_path:
            self.info(msgs, self.compare_with(os.path.normpath(actual_path),
                                              expected_path))
        elif expected_path:
            self.info(msgs, 'Expected file %s' % expected_path)
        elif actual_path:
            self.info(msgs, 'Actual file %s' % os.path.normpath(actual_path))
        else:
            self.info(msgs, 'No files')
        if ignore_substrings or ignore_patterns:
            self.info(msgs, 'Note exclusions:')
        if ignore_substrings:
            for pattern in ignore_substrings:
                self.info(msgs, '    ' + pattern)
        if ignore_patterns:
            for pattern in ignore_patterns:
                self.info(msgs, '    ' + pattern)
        if preprocess and actual_path:
            actualFilename = os.path.split(actual_path)[1]
            modifiedActual = os.path.join(self.tmp_dir,
                                          'actual-' + actualFilename)
            modifiedRef = os.path.join(self.tmp_dir,
                                       'expected-' + actualFilename)
            with open(modifiedActual, 'w') as f:
                f.write(actual if type(actual) == str
                        else '\n'.join(actual))
            with open(modifiedRef, 'w') as f:
                f.write('\n'.join(expected))
            self.info(msgs, self.compare_with(modifiedActual, modifiedRef,
                                              qualifier='preprocessed'))
        elif expected_path and not actual_path:
            expectedFilename = os.path.split(expected_path)[1]
            tmpActualFilename = os.path.join(self.tmp_dir,
                                            'actual-' + expectedFilename)
            with open(tmpActualFilename, 'w') as f:
                f.write(actual if type(actual) == str
                        else '\n'.join(actual))
            self.info(msgs,
                      self.compare_with(tmpActualFilename, expected_path))




# elements_verify_92.py

from __future__ import print_function
import pandas as pd

from tdda.constraints.pd.constraints import verify_df

df = pd.read_csv('testdata/elements92.csv')
print(verify_df(df, 'elements92.tdda'))



"""
Test-Driven Data Analysis (Python TDDA library)
===============================================

The tdda package provides Python support for test-driven data analysis.

The tdda.referencetest library is used to support the creation of
reference tests, based on either unittest or pytest.

For usage details:

    >>> from tdda import referencetest
    >>> help(referencetest)


The tdda.constraints library is used to 'discover' constraints
from a (Pandas) DataFrame, write them out as JSON, and to verify that
datasets meet the constraints in the constraints file.

For usage details:

    >>> from tdda import constraints
    >>> help(constraints)


The tdda package also includes rexpy, a tool for automatically
inferring regular expressions from a single field of data examples.

For usage details:

    >>> from tdda import rexpy
    >>> help(rexpy)

"""
__version__ = '0.5.12'



# -*- coding: utf-8 -*-

#
# We use regex in preference to re when available.
# It has better handling for international character sets.
#
# If you use
#
#     from tdda.rexpy.relib import re, relib
#
# then relib will allow you to know which library is in use.
#


# try:
#     import regex as re
#     relib = 'regex'
# except ImportError:
if 1:
    import re
    relib = 're'







# -*- coding: utf-8 -*-

"""
Extensions to the ``tdda`` command line tool, to support databases.
"""

from __future__ import print_function

import sys

from tdda.constraints.extension import ExtensionBase

from tdda.constraints.db.drivers import applicable
from tdda.constraints.db.discover import DatabaseDiscoverer
from tdda.constraints.db.verify import DatabaseVerifier
from tdda.constraints.db.detect import DatabaseDetector


class TDDADatabaseExtension(ExtensionBase):
    def __init__(self, argv, verbose=False):
        ExtensionBase.__init__(self, argv, verbose=verbose)

    def applicable(self):
        return applicable(self.argv)

    def help(self, stream=sys.stdout):
        print('  - Tables from PostgreSQL databases ("postgres:table-name")',
              file=stream)
        print('  - Tables from MySQL databases ("mysql:table-name")',
              file=stream)
        print('  - Tables from SQLite databases ("sqlite:table-name")',
              file=stream)
        print('  - Collections from MongoDB NoSQL databases '
              '("mongodb:collection-name")', file=stream)

    def spec(self):
        return 'DBTYPE:tablename, or -dbtype DBTYPE and a database table'

    def discover(self):
        return DatabaseDiscoverer(self.argv, verbose=self.verbose).discover()

    def verify(self):
        return DatabaseVerifier(self.argv, verbose=self.verbose).verify()

    def detect(self):
        return DatabaseDetector(self.argv, verbose=self.verbose).detect()




# -*- coding: utf-8 -*-
"""
dataframes.py: Trivial result generation functions for illustrating
               tdda.referencetest with Pandas Dataframes.

Source repository: http://github.com/tdda/tdda

License: MIT

Copyright (c) Stochastic Solutions Limited 2016-2018
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import pandas as pd


def generate_dataframe(nrows=10, precision=3):
    """
    Generate a simple Pandas DataFrame with examples of integer, real and
    string columns.
    """
    df = pd.DataFrame({'i': range(nrows)})
    df['r'] = df.i * (10/9)
    df['random'] = np.random.randint(0, nrows-1, nrows)
    df['b'] = df.i >= nrows/2
    df['i_square'] = df.i * df.i
    df['r_square'] = df.r * df.r
    df['s'] = 'result ' + df.r_square.round(precision).astype(str)
    return df

